Require Export Metatheory.

Definition label := atom.
Definition ptr := atom.
Definition labelset := atoms.
Definition lset_subset s s' := forall (a:atom), a `in` s -> a `in` s'.

Lemma subset_transitive: forall s1 s2 s3,
  lset_subset s1 s2 -> lset_subset s2 s3 -> lset_subset s1 s3.
Proof.
  unfold lset_subset in *. intros. apply H0. apply H. assumption.
Qed.

Lemma set_union_assoc : forall s1 s2 s3,
  s1 `union` (s2 `union` s3) = (s1 `union` s2) `union` s3.
Proof.
  intros. decideFSet.
Qed.

Lemma subseteq : forall s1 s2,
  lset_subset s1 s2 -> lset_subset s2 s1 -> s1 = s2.
Proof.
Admitted.

Lemma in_union : forall x s1 s2,
  x `in` (s1 `union` s2) ->
  x `in` s1 \/ x `in` s2.
Proof.
  intros. decideFSet.
Qed.

Lemma subset_add : forall s1 s2 s3 s4,
  lset_subset s1 s2 ->
  lset_subset s3 s4 ->
  lset_subset (s1 `union` s3) (s2 `union` s4).
Proof.
  unfold lset_subset.
  intros.
  destruct (in_union a s1 s3 H1).
  assert (a `in` s2). apply (H a H2). decideFSet.
  assert (a `in` s4). apply (H0 a H2). decideFSet.
Qed.

Lemma subset_refl : forall s, lset_subset s s.
Proof.
  intros. unfold lset_subset. intros. assumption.
Qed.

Lemma union_subset : forall s1 s2 s,
  lset_subset (s1 `union` s2) s ->
  lset_subset s1 s /\ lset_subset s2 s.
Proof.
  intros. unfold lset_subset in *. split; intros; apply H; decideFSet.
Qed.

Hint Resolve subset_transitive.

Inductive phi : Set :=
  | make_phi : labelset -> labelset -> labelset -> phi.

Function phi_fld_fst (p: phi) : labelset :=
  match p with
    | make_phi a e o => a
  end.

Function phi_fld_rst (p: phi) : labelset :=
  match p with
    | make_phi a e o => o
  end.

Function phi_fld_eff (p: phi) : labelset :=
  match p with
    | make_phi a e o => e
  end.

Inductive comb_cflow : phi -> phi -> phi -> Prop :=
  | make_comb : forall (fst1 teff1 teff2 rst2: labelset),
      comb_cflow
      (make_phi fst1 teff1 (rst2 `union` teff2))
      (make_phi (fst1 `union` teff1) teff2 rst2)
      (make_phi fst1 (teff1 `union` teff2) rst2).

Inductive phi_sub : phi -> phi -> Prop :=
  | sub_context : forall (fst1 fst2 teff1 teff2 rst1 rst2: labelset),
      lset_subset fst2 fst1 ->
      lset_subset teff1 teff2 ->
      lset_subset rst2 rst1 ->
      phi_sub
        (make_phi fst1 teff1 rst1)
        (make_phi fst2 teff2 rst2).

Lemma phi_sub_eq : forall p1 p2,
  phi_sub p1 p2 -> phi_sub p2 p1 -> p1 = p2.
Proof.
  intros. induction H. inversion H0. subst. f_equal; apply subseteq; assumption.
Qed.

Lemma phi_sub_refl : forall p, phi_sub p p.
Proof.
  intros. destruct p. apply sub_context; apply subset_refl.
Qed.

Lemma phi_sub_trans : forall p1 p2 p3, phi_sub p1 p2 -> phi_sub p2 p3 -> phi_sub p1 p3.
Proof.
  intros p1 p2 p3 H1 H2. inversion H1. subst. inversion H2. subst. apply sub_context.
  apply subset_transitive with (s2 := fst2). assumption. assumption.
  apply subset_transitive with (s2 := teff2). assumption. assumption.
  apply subset_transitive with (s2 := rst2). assumption. assumption.
Qed.

Lemma phi_comb_teff : forall p1 p2 p3,
  comb_cflow p1 p2 p3 ->
  lset_subset (phi_fld_eff p1) (phi_fld_eff p3) /\
  lset_subset (phi_fld_eff p2) (phi_fld_eff p3).
Proof.
  intros. destruct H. simpl. unfold lset_subset; split; decideFSet.
Qed.

Parameter full_set : labelset.
Axiom everything_in_full_set : forall x, x `in` full_set.

Definition phi_0 := make_phi full_set {} full_set.

Lemma subset_full_set : forall s, lset_subset s full_set.
Proof.
  unfold lset_subset. intros. apply everything_in_full_set.
Qed.

Lemma full_set_subset : forall s,
  lset_subset full_set s -> s = full_set.
Proof.
  intros. apply subseteq. apply subset_full_set. assumption.
Qed.

Lemma union_full_set : forall s,
  full_set `union` s = full_set.
Proof.
  intros.
  apply subseteq. apply subset_full_set.
  unfold lset_subset.
  intros.
  decideFSet.
Qed.

Lemma empty_set_subset : forall s, lset_subset {} s.
Proof.
  unfold lset_subset. intros. decideFSet.
Qed.

Lemma id_subset : forall s, lset_subset s s.
Proof.
  unfold lset_subset. intros. assumption.
Qed.

Lemma phi_0_sub_all : forall p, phi_sub phi_0 p.
Proof.
  intros.
  destruct p. unfold phi_0.
  apply sub_context; auto using subset_full_set, empty_set_subset.
Qed.

Inductive typ : Set :=
  | typ_int : typ
  | typ_ref : labelset -> typ -> typ
  | typ_fun : phi -> typ -> typ -> typ.

Inductive subtype : typ -> typ -> Prop :=
  | sub_int : subtype typ_int typ_int
  | sub_ref : forall t t' teff teff',
      subtype t t' ->
      subtype t' t ->
      lset_subset teff teff' ->
      subtype (typ_ref teff t) (typ_ref teff' t')
  | sub_fun : forall t1 t1' t2 t2' p p',
      subtype t1' t1 ->
      subtype t2 t2' ->
      phi_sub p p' ->
      subtype (typ_fun p t1 t2) (typ_fun p' t1' t2').

Hint Constructors subtype.

Lemma subeq : forall t1 t2,
  subtype t1 t2 -> subtype t2 t1 -> t1 = t2.
Proof.
  intros. induction H.
  reflexivity.
  f_equal. apply subseteq. assumption. inversion H0.
    assumption.
    apply IHsubtype1. assumption.
  inversion H0. subst.
  f_equal. apply phi_sub_eq; assumption.
    assert (t1' = t1). apply IHsubtype1. assumption. subst. reflexivity.
    apply IHsubtype2. assumption.
Qed.

Lemma sub_refl : forall s, subtype s s.
Proof.
  intros. induction s.
    apply sub_int.
    apply sub_ref. assumption. assumption. apply subset_refl.
    apply sub_fun. assumption. assumption. apply phi_sub_refl.
Qed.

Inductive exp : Set :=
  | e_nat : nat -> exp
  | e_bvar : nat -> exp
  | e_fvar : atom -> exp
  | e_lam : exp -> exp
  | e_app : exp -> exp -> exp
(*  | e_let : exp -> exp -> exp *)
(*  | e_if : exp -> exp -> exp -> exp *)
  | e_ref : label -> exp -> exp
  | e_deref : exp -> exp
  | e_assign : exp -> exp -> exp
  | e_loc : ptr -> label -> exp
.

Inductive value : exp -> Prop :=
  | int_value : forall n, value (e_nat n)
  | lam_value : forall e, value (e_lam e)
  | loc_value : forall ptr label, value (e_loc ptr label).

Hint Constructors value.

Coercion e_bvar : nat >-> exp.
Coercion e_fvar : atom >-> exp.

Fixpoint subst (z : atom) (u : exp) (e : exp) {struct e} : exp :=
  match e with
    | e_nat n => e_nat n
    | e_bvar i => e_bvar i
    | e_fvar x => if x == z then u else (e_fvar x)
    | e_lam e1 => e_lam (subst z u e1)
    | e_app e1 e2 => e_app (subst z u e1) (subst z u e2)
(*    | e_let e1 e2 => e_let (subst z u e1) (subst z u e2) *)
(*    | e_if e1 e2 e3 => e_if (subst z u e1) (subst z u e2) (subst z u e3) *)
    | e_ref teff e => e_ref teff (subst z u e)
    | e_deref e => e_deref (subst z u e)
    | e_assign e1 e2 => e_assign (subst z u e1) (subst z u e2)
    | e_loc ptr label => e_loc ptr label
  end.

Notation "[ z ~> u ] e" := (subst z u e) (at level 68).

Fixpoint fv (e : exp) {struct e} : atoms :=
  match e with
    | e_nat n => {}
    | e_bvar i => {}
    | e_fvar x => singleton x
    | e_lam e1 => fv e1
    | e_app e1 e2 => (fv e1) `union` (fv e2)
(*    | e_let e1 e2 => (fv e1) `union` (fv e2) *)
(*    | e_if e1 e2 e3 => (fv e1) `union` (fv e2) `union` (fv e3) *)
    | e_ref teff e => fv e
    | e_deref e => fv e
    | e_assign e1 e2 => (fv e1) `union` (fv e2)
    | e_loc ptr label => singleton ptr
  end.

Fixpoint open_rec (k : nat) (u : exp) (e : exp) {struct e} : exp :=
  match e with
    | e_nat n => e_nat n
    | e_bvar i => if k === i then u else (e_bvar i)
    | e_fvar x => e_fvar x
    | e_lam e1 => e_lam (open_rec (S k) u e1)
    | e_app e1 e2 => e_app (open_rec k u e1) (open_rec k u e2)
(*    | e_let e1 e2 => e_let (open_rec k u e1) (open_rec (S k) u e2) *)
(*    | e_if e1 e2 e3 => e_if (open_rec k u e1) (open_rec k u e2) (open_rec k u e3) *)
    | e_ref teff e => e_ref teff (open_rec k u e)
    | e_deref e => e_deref (open_rec k u e)
    | e_assign e1 e2 => e_assign (open_rec k u e1) (open_rec k u e2)
    | e_loc ptr label => e_loc ptr label
  end.

Notation "{ k ~> u } t" := (open_rec k u t) (at level 67).

Definition open e u := open_rec 0 u e.

Ltac gather_atoms :=
  let A := gather_atoms_with (fun x : atoms => x) in
  let B := gather_atoms_with (fun x : atom => singleton x) in
  let C := gather_atoms_with (fun x : list (atom * typ) => dom x) in
  let D := gather_atoms_with (fun x : exp => fv x) in
  constr:(A `union` B `union` C `union` D).

Tactic Notation "pick" "fresh" ident(x) :=
  let L := gather_atoms in
  (pick fresh x for L).

Tactic Notation
      "pick" "fresh" ident(atom_name) "and" "apply" constr(lemma) :=
  let L := gather_atoms in
  pick fresh atom_name excluding L and apply lemma.

Notation env := (list (atom * typ)).

Inductive typing : phi -> env -> env -> exp -> typ -> Prop :=
  | typing_nat : forall n E P,
      ok E -> ok P -> typing phi_0 E P (e_nat n) typ_int
  | typing_var : forall E P (x : atom) T,
      ok E -> ok P ->
      binds x T E ->
      typing phi_0 E P (e_fvar x) T
  | typing_lam : forall L p E P e T1 T2,
      (forall x : atom, x `notin` L ->
        typing p ((x, T1) :: E) P (open e x) T2) ->
      typing phi_0 E P (e_lam e) (typ_fun p T1 T2)
  | typing_app : forall p1 p2 pf p E P e1 e2 T1 T2,
      typing p1 E P e1 (typ_fun pf T1 T2) ->
      typing p2 E P e2 T1 ->
      (exists p3, comb_cflow p1 p2 p3 /\ comb_cflow p3 pf p) ->
      typing p E P (e_app e1 e2) T2
(*  | typing_let : forall p1 p2 T1 L p E P e1 e2 T2,
      typing p1 E P e1 T1 ->
      (forall x : atom, x `notin` L ->
        typing p2 ((x, T1)::E) P (open e2 x) T2) ->
      comb_cflow p1 p2 p ->
      typing p E P (e_let e1 e2) T2 *)
(*  | typing_if : forall p1 p2 p E P e1 e2 e3 T,
      typing p1 E P e1 typ_int ->
      typing p2 E P e2 T ->
      typing p2 E P e3 T ->
      comb_cflow p1 p2 p ->
      typing p E P (e_if e1 e2 e3) T *)
  | typing_sub : forall p p' E P e T T',
      typing p' E P e T' ->
      subtype T' T ->
      phi_sub p' p ->
      typing p E P e T
  | typing_ref : forall p E P e t label,
      typing p E P e t ->
      typing p E P (e_ref label e) (typ_ref (singleton label) t)
  | typing_deref : forall p p1 E P e teff t fst rst,
      typing p1 E P e (typ_ref teff t) ->
      comb_cflow p1 (make_phi fst teff rst) p ->
      typing p E P (e_deref e) t
  | typing_assign : forall p1 p2 fst rst teff p p' E P e1 e2 t,
      typing p1 E P e1 (typ_ref teff t) ->
      typing p2 E P e2 t ->
      comb_cflow p1 p2 p' ->
      comb_cflow p' (make_phi fst teff rst) p ->
      typing p E P (e_assign e1 e2) t
  | typing_loc : forall E P ptr label t,
      ok E -> ok P ->
      binds ptr t P ->
      typing phi_0 E P (e_loc ptr label) (typ_ref (singleton label) t)  
.

Hint Constructors typing.

Lemma typing_env_ok : forall p E P e t,
  typing p E P e t -> ok E /\ ok P.
Proof.
  intros. induction H; auto.
  pick fresh x. destruct (H0 x). auto. inversion H1. auto.
Qed.

Notation heap := (list (ptr * exp)).

Fixpoint replace (h: heap) (r: ptr) (v: exp) {struct h} :=
  match h with
    | nil => nil
    | (r',v')::tl =>
        if r' == r then cons (r, v) tl
        else (r', v')::(replace tl r v)
  end.

Lemma binds_after_replace : forall (h: heap) (r: ptr) (v v': exp),
  binds r v h -> binds r v' (replace h r v').
Proof.
  intros. induction h as [|(l,e)].
  Case "nil".
    unfold binds in H. inversion H.
  Case "cons".
    unfold binds. simpl. destruct (l == r). 
    Case "equal". 
      simpl. destruct (r == r). reflexivity.
        destruct n. reflexivity.
    Case "different".
      simpl. destruct (r == l).
        subst. destruct n. reflexivity.
        unfold binds in *. simpl in *. apply IHh. destruct (r == l).
          subst. destruct n. reflexivity. apply H.
Qed.

Lemma binds_other_after_replace : forall h r r' v v',
  binds r v h -> r <> r' -> binds r v (replace h r' v').
Proof.
  intros. induction h as [|(l,e)].
  Case "nil". unfold binds in H. inversion H.
  Case "cons".
    unfold binds. simpl. destruct (l==r).
    Case "equal". subst.
      destruct (r == r'). destruct H0. assumption.
        simpl. unfold binds in H. simpl in H.
        destruct (r == r). assumption.
      unfold binds in *. apply IHh. assumption.
    Case "different".
      unfold binds in *. simpl in *.
      destruct (r == l). destruct n; auto. simpl.
      destruct (l == r'). subst. simpl. destruct (r == r'). destruct n. auto.
      unfold binds in *. simpl in *. destruct (r == r'). destruct n. auto.
      assumption.
      simpl. destruct (r ==l). destruct n; auto.
      apply IHh. assumption.
Qed.

Inductive result : Set :=
  | exp_result : exp -> result
  | err_result : result.

Inductive evaluation :
  labelset -> labelset -> heap -> exp ->
  labelset ->
  labelset -> labelset -> heap -> result -> Prop :=
  | id : forall fst rst heap v,
      value v ->
      evaluation fst rst heap v {} fst rst heap (exp_result v)
  | call : forall e1 e2 e v2 v fst fst1 fst2 fst' rst rst1 rst2 rst'
                  heap heap1 heap2 heap' teff1 teff2 teff3,
      evaluation fst rst heap e1 teff1 fst1 rst1 heap1 (exp_result (e_lam e)) ->
      evaluation fst1 rst1 heap1 e2 teff2 fst2 rst2 heap2 (exp_result v2) ->
      evaluation fst2 rst2 heap2 (open e v2) teff3 fst' rst' heap' (exp_result v) ->
      evaluation fst rst heap (e_app e1 e2) (teff1 `union` teff2 `union` teff3)
        fst' rst' heap' (exp_result v)
  | ref : forall fst fst' rst rst' heap heap' e teff v r l,
      evaluation fst rst heap e teff fst' rst' heap' (exp_result v) ->
      r `notin` (dom heap') ->
      evaluation fst rst heap (e_ref l e) teff fst' rst' ((r,v)::heap')
        (exp_result (e_loc r l))
  | deref : forall fst fst' rst rst' teff label heap heap' e ptr v,
      evaluation fst rst heap e teff fst'
        (rst' `union` (singleton label)) heap' (exp_result (e_loc ptr label)) ->
      ptr `in` (dom heap') ->
      binds ptr v heap' ->
      evaluation fst rst heap (e_deref e) (teff `union` (singleton label))
        (fst' `union` (singleton label)) rst' heap' (exp_result v)
  | assign : forall fst fst1 fst2 teff1 teff2 rst rst1 rst2
                    e1 e2 heap heap1 heap2 ptr label v v',
      evaluation fst rst heap e1 teff1
                 fst1 rst1 heap1 (exp_result (e_loc ptr label)) ->
      evaluation fst1 rst1 heap1 e2 teff2
                 fst2 (rst2 `union` (singleton label)) heap2 (exp_result v) ->
      binds ptr v' heap2 ->
      evaluation fst rst heap (e_assign e1 e2)
                 ((teff1 `union` teff2) `union` (singleton label))
                 (fst2 `union` (singleton label)) rst2
                 (replace heap2 ptr v) (exp_result v)
(*  | if_t : forall fst rst heap e1 e2 e3 v fst1 fst2 rst1 rst2 heap1 heap2 teff1 teff2,
      evaluation fst rst heap e1 teff1 fst1 rst1 heap1 (exp_result (e_nat 0)) ->
      evaluation fst1 rst1 heap1 e2 teff2 fst2 rst2 heap2 (exp_result v) ->
      evaluation fst rst heap (e_if e1 e2 e3) (teff1 `union` teff2) fst2 rst2 heap2
        (exp_result v)
  | if_f : forall fst rst heap e1 e2 e3 v n fst1 fst2 rst1 rst2 heap1 heap2 teff1 teff2,
      evaluation fst rst heap e1 teff1 fst1 rst1 heap1 (exp_result (e_nat n)) ->
      n <> 0 ->
      evaluation fst1 rst1 heap1 e3 teff2 fst2 rst2 heap2 (exp_result v) ->
      evaluation fst rst heap (e_if e1 e2 e3) (teff1 `union` teff2) fst2 rst2 heap2
        (exp_result v) *)
  | call_w : forall fst rst heap e1 e2 teff fst' rst' heap' v,
      evaluation fst rst heap e1 teff fst' rst' heap' (exp_result v) ->
      (forall e, v <> (e_lam e)) ->
      evaluation fst rst heap (e_app e1 e2) teff fst' rst' heap err_result
(*  | if_w : forall fst rst heap e1 e2 e3 teff fst' rst' heap' v1,
      evaluation fst rst heap e1 teff fst' rst' heap' (exp_result v1) ->
      (forall n, v1 <> (e_nat n)) ->
      evaluation fst rst heap (e_if e1 e2 e3) teff fst' rst' heap err_result *)
  | deref_h_w : forall fst rst e heap teff fst' rst' heap' ptr label,
      evaluation fst rst heap e teff fst' rst' heap'
        (exp_result (e_loc ptr label)) ->
      ptr `notin` (dom heap') ->
      evaluation fst rst heap (e_deref e) teff fst' rst' heap err_result
  | deref_l_w : forall fst rst e heap teff fst' rst' heap' ptr label,
      evaluation fst rst heap e teff fst' rst' heap'
        (exp_result (e_loc ptr label)) ->
      ptr `in` (dom heap') ->
      ~(label `in` fst') ->
      evaluation fst rst heap (e_deref e) teff fst' rst' heap err_result
  | assign_h_w : forall fst rst heap e1 teff1 fst1 rst1 heap1 ptr label
                        e2 teff2 fst2 rst2 heap2 v,
      evaluation fst rst heap e1 teff1
                 fst1 rst1 heap1 (exp_result (e_loc ptr label)) ->
      evaluation fst1 rst1 heap1 e2 teff2
                 fst2 rst2 heap2 (exp_result v) ->
      ptr `notin` (dom heap2) ->
      evaluation fst rst heap (e_assign e1 e2) (teff1 `union` teff2)
                 fst2 rst2 heap err_result
  | assign_l_w : forall fst rst heap e1 teff1 fst1 rst1 heap1 ptr label
                        e2 teff2 fst2 rst2 heap2 v' v,
      evaluation fst rst heap e1 teff1 fst1 rst1 heap1
                 (exp_result (e_loc ptr label)) ->
      evaluation fst1 rst1 heap1 e2 teff2 fst2 rst2 heap2 (exp_result v) ->
      binds ptr v' heap2 ->
      label `notin` rst2 ->
      evaluation fst rst heap (e_assign e1 e2) (teff1 `union` teff2) fst2 rst2 heap err_result
.

Hint Constructors evaluation.

Tactic Notation "rewrite" "set" constr(s) "as" constr(X) :=
  let H := fresh "H" in
  assert (s = X) as H; [ decideFSet | rewrite H; clear H ].

Lemma evaluation_weakening : forall fst rst heap e eff fst' rst' heap' v fstw rstw,
  evaluation fst rst heap e eff fst' rst' heap' (exp_result v) ->
  evaluation (fst `union` fstw) (rst `union` rstw) heap e eff
    (fst' `union` fstw) (rst' `union` rstw) heap' (exp_result v).
Proof.
  intros.
  remember (exp_result v) as R.
  generalize dependent v.
  generalize dependent fstw.
  generalize dependent rstw.
  induction H;
    intros rstw fstw vv Eq; inversion Eq; subst; clear Eq; eauto.
  Case "deref".
    intros.
    rewrite set ((fst' `union` AtomSet.F.singleton label0) `union` fstw)
      as ((fst' `union` fstw) `union` AtomSet.F.singleton label0).
    eapply deref.
    rewrite set ((rst' `union` rstw) `union` AtomSet.F.singleton label0)
      as ((rst' `union` AtomSet.F.singleton label0) `union` rstw).
    eapply IHevaluation.
    reflexivity. assumption. assumption.
  Case "assign".
    rewrite set ((fst2 `union` AtomSet.F.singleton label0) `union` fstw)
      as ((fst2 `union` fstw) `union` AtomSet.F.singleton label0).
    eapply assign.
    eapply IHevaluation1; auto.
    rewrite set ((rst2 `union` rstw) `union` AtomSet.F.singleton label0)
      as ((rst2 `union` AtomSet.F.singleton label0) `union` rstw).
    eapply IHevaluation2; auto.
    apply H1.
Qed.

Lemma forward_effect_subset : forall fst rst fst' rst' teff e v H H',
  evaluation fst rst H e teff fst' rst' H' (exp_result v) ->
  lset_subset fst fst'.
Proof.
  intros.
  induction H0; eauto using subset_refl, subset_transitive.
  rewrite set fst as (fst `union` {});
  apply subset_add; try assumption; apply empty_set_subset.
  rewrite set fst as (fst `union` {}).
  apply subset_add; eauto using subset_transitive, empty_set_subset.
Qed.

Lemma backward_effect_subset : forall fst rst fst' rst' teff e v H H',
  evaluation fst rst H e teff fst' rst' H' (exp_result v) ->
  lset_subset rst' rst.
Proof.
  intros.
  induction H0; eauto using subset_refl, subset_transitive.
  destruct (union_subset rst' (singleton label0) rst IHevaluation).
  assumption.
  destruct (union_subset rst2 (singleton label0) rst1 IHevaluation2).
  apply subset_transitive with rst1; assumption.
Qed.

Lemma adequacy : forall fst rst heap e eff fst' rst' heap' v,
  value v ->
  evaluation fst rst heap e eff fst' rst' heap' (exp_result v) ->
  fst' = fst `union` eff /\ rst = rst' `union` eff.
Proof.
  intros fst rst heap e eff fst' rst' heap' v V H.
  split; induction H; try assumption; try decideFSet.
Qed.

Ltac propagate_full_set f :=
  assert (f = full_set);
    [ apply full_set_subset; eauto using forward_effect_subset, backward_effect_subset
      | subst f ].

Lemma canonical_evaluation : forall fst rst heap e eff fst' rst' heap' v,
  fst = full_set ->
  rst = full_set ->
  fst' = full_set ->
  rst' = full_set ->
  evaluation fst rst heap e eff fst' rst' heap' (exp_result v) ->
  evaluation {} eff heap e eff eff {} heap' (exp_result v).
Proof.
  intros.
  remember (exp_result v) as R.
  generalize dependent v.
  induction H3; intros vv Eq; inversion Eq; subst vv; clear Eq; eauto;
    subst fst rst.
  Case "call".
    intros.
    subst fst' rst'.
    propagate_full_set fst1.
    propagate_full_set fst2.
    propagate_full_set rst2.
    propagate_full_set rst1.
    eapply call.
    Subcase "e1".
      rewrite set {} as ({} `union` {}).
      apply evaluation_weakening. eauto.
    Subcase "e2".
      rewrite set (teff1 `union` {}) as ({} `union` teff1).
      rewrite set ({} `union` teff2 `union` teff3) as (teff2 `union` teff3).
      apply evaluation_weakening. eauto.      
    Subcase "e3".
      rewrite set {} as ({} `union` {}).
      rewrite set (teff2 `union` teff1) as ({} `union` (teff1 `union` teff2)).
      rewrite set (({} `union` {}) `union` teff3) as (teff3 `union` {}).
      rewrite set (teff1 `union` teff2 `union` teff3) as
        (teff3 `union` teff1 `union` teff2).
      apply evaluation_weakening. eauto. 
  Case "deref".
    subst rst'.
    propagate_full_set fst'.
    eapply deref.
    rewrite set {} as ({} `union` {}).
    rewrite set teff as (teff `union` {}).
    rewrite set (({} `union` {}) `union` AtomSet.F.singleton label0)
      as ({} `union` AtomSet.F.singleton label0).
    rewrite set ((teff `union` {}) `union` AtomSet.F.singleton label0)
      as (teff `union` AtomSet.F.singleton label0).
    apply evaluation_weakening.
    rewrite set (teff `union` {}) as teff.
    eauto. assumption. assumption.
  Case "assign".
    subst rst2.
    propagate_full_set fst1.
    propagate_full_set fst2.
    propagate_full_set rst1.
    eapply assign; eauto.
    rewrite set {} as ({} `union` {}).
    rewrite set ((teff1 `union` teff2) `union` (singleton label0))
      as (teff1 `union` teff2 `union` (singleton label0)).
    apply evaluation_weakening. eauto.
    rewrite set ({} `union` teff2 `union` singleton label0)
      as (teff2 `union` singleton label0).
    rewrite set (teff1 `union` {}) as ({} `union` teff1).
    rewrite set (teff1 `union` teff2) as (teff2 `union` teff1).
    apply evaluation_weakening.
    eauto.
Qed.

Definition heap_ok (P:env) h := ok P /\ (dom h) = (dom P) /\
  forall r, r `in` (dom h) ->
  exists t, exists v,
    value v /\ binds r v h /\ binds r t P /\ typing phi_0 nil P v t.

Lemma empty_heap_ok : heap_ok nil nil.
Proof.
  split. apply ok_nil.
  split. reflexivity.
  intros. decideFSet.
Qed.

(*
Definition env_subset E E' := forall (x:atom) (t:typ),
  binds x t E -> binds x t E'.
*)

Lemma typing_weakening_str : forall P P1 P2 E F G p e t,
  typing p (G ++ E) (P1 ++ P) e t ->
  ok (G ++ F ++ E) ->
  ok (P1 ++ P2 ++ P) ->
  typing p (G ++ F ++ E) (P1 ++ P2 ++ P) e t.
Proof.
  intros P P1 P2 E F G p e t H.
  remember (G ++ E) as E'.
  remember (P1 ++ P) as P'.
  generalize dependent G.
  generalize dependent P1.
  induction H; intros P1 Eqp G EqE OkE OkP; subst; eauto.
  Case "lam".
    pick fresh x and apply typing_lam. apply H0 with (G0 := (x,T1)::G); auto.
    rewrite cons_concat_assoc. auto.
Qed.

Lemma typing_weakening : forall p E F P P1 e T,
    typing p E P e T ->
    ok (F ++ E) ->
    ok (P1 ++ P) ->
    typing p (F ++ E) (P1 ++ P) e T.
Proof.
  intros p E F P P1 e T H J.
  rewrite <- (nil_concat _ (F ++ E)).
  rewrite <- (nil_concat _ (P1 ++ P)).
  apply typing_weakening_str; auto.
Qed.

Lemma env_ok : forall H E, heap_ok E H -> ok E.
Proof.
  intros.
  unfold heap_ok in H0. destruct H0. auto.
Qed.

Lemma typing_subst_var_case : forall p E F P u S T z x,
  binds x T (F ++ (z, S) :: E) ->
  ok (F ++ (z, S) :: E) ->
  ok P ->
  typing p E P u S ->
  typing p (F ++ E) P ([z ~> u] x) T.
Proof.
  intros p E F P u S T z x H J Pok K.
  simpl.
  destruct (x == z).
  Case "x = z".
    subst.
    assert (T = S).
      eapply binds_mid_eq_cons. apply H. apply J.
    subst.
    rewrite <- (nil_concat _ P).
    apply typing_weakening.
      apply K.
      eapply ok_remove_mid_cons. apply J. simpl. auto.
  Case "x <> z".
    apply typing_sub with (p' := phi_0) (T' := T).
    apply typing_var.
      apply ok_remove_mid_cons with (x := z) (a := S). apply J. auto.
      apply binds_remove_mid_cons with (y := z) (b := S). apply H. apply n.
    apply sub_refl. apply phi_0_sub_all.
Qed.

Inductive lc : exp -> Prop :=
  | lc_nat : forall n, lc (e_nat n)
  | lc_fvar : forall x,
      lc (e_fvar x)
  | lc_lam : forall L e,
      (forall x:atom, x `notin` L -> lc (open e x)) ->
      lc (e_lam e)
  | lc_app : forall e1 e2,
      lc e1 ->
      lc e2 ->
      lc (e_app e1 e2)
(*  | lc_let : forall L e1 e2,
      lc e1 ->
      (forall x:atom, x `notin` L -> lc (open e2 x)) -> lc (e_let e1 e2) *)
(*  | lc_if : forall e1 e2 e3,
      lc e1 -> lc e2 -> lc e3 -> lc (e_if e1 e2 e3) *)
  | lc_ref : forall teff e,
      lc e -> lc (e_ref teff e)
  | lc_deref : forall e,
      lc e -> lc (e_deref e)
  | lc_assign : forall e1 e2,
      lc e1 -> lc e2 -> lc (e_assign e1 e2)
  | lc_loc : forall ptr label, lc (e_loc ptr label)
.

Hint Constructors lc.

Lemma typing_regular_lc : forall p E P e T,
  typing p E P e T -> lc e.
Proof.
  intros p E P e T H. induction H; eauto.
Qed.

Lemma open_rec_lc_core : forall e j v i u,
  i <> j ->
  {j ~> v} e = {i ~> u} ({j ~> v} e) ->
  e = {i ~> u} e.
Proof with eauto*.
  induction e; intros j v i u Neq H; simpl in *;
      try solve [inversion H; f_equal; eauto].
  Case "bvar".
    destruct (j === n)...
    destruct (i === n)...
Qed.

Lemma open_rec_lc : forall k u e,
  lc e ->
  e = {k ~> u} e.
Proof.
  intros k u e LC.
  generalize dependent k.
  induction LC; simpl; intro k; f_equal; auto.
  Case "lc_lam".
    unfold open in *.
    pick fresh x for L.
    apply open_rec_lc_core with (i := S k) (j := 0) (u := u) (v := x). auto.
    auto.
Qed.

Lemma subst_open_rec : forall e1 e2 u x k,
  lc u ->
  [x ~> u] ({k ~> e2} e1) = {k ~> [x ~> u] e2} ([x ~> u] e1).
Proof.
  induction e1; intros e2 u x k H; simpl in *; f_equal; auto.
  Case "bvar".
    destruct (k === n); auto.
  Case "fvar".
    destruct (a == x); subst; auto.
    apply open_rec_lc; auto.
Qed.

Lemma subst_open_var : forall (x y : atom) u e,
  y <> x ->
  lc u ->
  open ([x ~> u] e) y = [x ~> u] (open e y).
Proof.
  intros x y u e Neq H.
  unfold open.
  rewrite subst_open_rec with (e2 := y).
  simpl.
  destruct (y == x).
    Case "y=x".
    destruct Neq. auto.
    Case "y<>x".
    auto.
  auto.
Qed.

Lemma typing_subst_strengthened : forall p E F P e u S T z,
  typing p (F ++ (z, S) :: E) P e T ->
  typing phi_0 E P u S ->
  typing p (F ++ E) P ([z ~> u] e) T.
Proof.
  intros p E F P e u S T z He Hu.
  remember (F ++ (z, S) :: E) as E'.
  generalize dependent F.
  induction He; intros F Eq; subst; simpl; eauto.
  Case "typing_var".
    apply (typing_subst_var_case phi_0 E F P u S T z x); auto.
  Case "typing_lam".
    pick fresh y and apply typing_lam.
    rewrite subst_open_var.
    rewrite <- cons_concat_assoc. auto. auto.
    apply (typing_regular_lc phi_0 E P u S Hu).
(*  Case "typing_let".
    pick fresh y and apply (typing_let p1 p2 T1). auto.
      rewrite subst_open_var.
      apply H0 with (x := y) (F0 := (y,T1)::F); auto. auto.
      apply (typing_regular_lc p E P u S (Hu p)). assumption. *)
Qed.

Lemma typing_subst : forall p E P e u S T z,
  typing p ((z, S) :: E) P e T ->
  typing phi_0 E P u S ->
  typing p E P ([z ~> u] e) T.
Proof.
  intros p E P e u S T z H J.
  rewrite <- (nil_concat _ E).
  eapply typing_subst_strengthened.
    rewrite nil_concat. apply H.
    apply J.
Qed.

Lemma generalize_value_phi : forall v E P p t,
  value v ->
  typing p E P v t ->
  typing phi_0 E P v t.
Proof with rewrite H2; eapply typing_sub;auto; eauto using phi_sub_refl.
  intros v E P p t V H.
  induction H; intros; inversion V; eauto...
Qed.

Lemma subst_fresh : forall (x : atom) e u,
  x `notin` fv e -> [x ~> u] e = e.
Proof.
  intros x e u H.
  induction e; subst; simpl in *; f_equal; auto.
  Case "fvar".
   destruct (a == x); subst; auto. decideFSet.
Qed.

Lemma subst_intro : forall (x : atom) u e,
  x `notin` (fv e) ->
  lc u ->
  open e u = [x ~> u](open e x).
Proof.
  intros x u e H J.
  unfold open.
  rewrite subst_open_rec; auto.
  simpl.
  destruct (x == x).
  Case "x = x".
    rewrite subst_fresh; auto.
  Case "x <> x".
    destruct n; auto.
Qed.

Definition transitivity_on Q := forall S T,
  subtype S Q -> subtype Q T -> subtype S T.

Lemma sub_transitivity : forall Q,
  transitivity_on Q.
Proof with simpl_env; auto.
  unfold transitivity_on.
  intros Q S T SsubQ QsubT.
  generalize dependent T.
  generalize dependent S.
  remember Q as Q' in |-.
  generalize dependent Q'.
  induction Q;
    intros Q' EQ S SsubQ;
    induction SsubQ; try discriminate;
    inversion EQ; subst;
    intros T QsubT;
    inversion QsubT; subst; auto.
  Case "ref".
    assert (t = Q). apply subeq; assumption.
    assert (t' = Q). apply subeq; assumption. subst.
    apply sub_ref; auto. apply subset_transitive with (s2 := l); assumption.
  Case "fun".
    apply sub_fun.
      apply IHQ1 with (Q' := Q1). reflexivity. assumption. assumption.
      apply IHQ2 with (Q' := Q2). reflexivity. assumption. assumption.
      apply phi_sub_trans with (p2 := p). assumption. assumption.
Qed.

Lemma ok_env_narrowing : forall (F E:env) x t t',
  ok (F ++ ((x,t)::nil) ++ E) -> ok (F ++ ((x,t')::nil) ++ E).
Proof.
  induction F; intros E x t t' H; inversion H; subst; simpl; apply ok_cons; auto.
  simpl_env in *. eapply IHF. apply H2. 
    simpl_env in *. assumption.
Qed.

Lemma sub_narrowing_str : forall x p t1 t2 F E P e t,
  subtype t1 t2 ->
  typing p (F ++ (x, t2)::E) P e t ->
  typing p (F ++ (x, t1)::E) P e t.
Proof with simpl_env in *; eauto using ok_env_narrowing, phi_sub_refl.
  intros x p t1 t2 F E P e t Sub T.
  remember (F ++ (x,t2)::E) as G.
  generalize dependent F.
  induction T; intros F Eq; subst...
  Case "var". destruct (x == x0).
    Subcase "x = x0".
    subst. binds_get H1. apply typing_sub with (T' := t1) (p' := phi_0)...
      apply typing_var; auto.
      apply ok_env_narrowing with (t := t2). assumption. 
      apply binds_mid. eapply ok_env_narrowing... 
    Subcase "x<>x0".
      apply typing_var; eauto using ok_env_narrowing.
      binds_cases H1; eauto using binds_tail, binds_head.      
  Case "lam".
    pick fresh y and apply typing_lam. apply H0 with (F0 := (y,T1)::F). auto.
    simpl_env. reflexivity.
(*  Case "let".
    pick fresh y and apply (typing_let p1 p2 T1).
      apply IHT...
      apply H0 with (F0 := (y,T1)::F); auto.
      assumption. *)
Qed.

Lemma sub_narrowing : forall p x E P e t1 t2 t,
  subtype t2 t1 ->
  typing p ((x, t1)::E) P e t ->
  typing p ((x, t2)::E) P e t.
Proof.
  intros.
  rewrite_env (nil ++ (x,t2)::E). eapply sub_narrowing_str. apply H.
  assumption.
Qed.

Lemma typing_nat_inv : forall p n E P t,
  typing p E P (e_nat n) t ->
    ok E /\ ok P /\ subtype typ_int t.
Proof.
  intros.
  destruct (typing_env_ok p E P (e_nat n) t H).
  split. assumption.
  split. assumption.
  remember (e_nat n) as e'.
  induction H; inversion Heqe'; subst.
  apply sub_refl.
  apply (sub_transitivity T'). auto. assumption.
Qed.

Lemma typing_var_inv : forall p E P x t,
  typing p E P (e_fvar x) t ->
  exists t', subtype t' t /\ ok E /\ ok P /\ binds x t' E.
Proof.
  intros. remember (e_fvar x) as e. induction H; inversion Heqe; subst.
  exists T. eauto using sub_refl.
  destruct (IHtyping H2) as [ t' [sub [ oke [ okp b ]]]]. 
  exists t'. split. apply (sub_transitivity T'); assumption. auto.
Qed.

Lemma typing_lam_inv : forall p E P e1 t,
  typing p E P (e_lam e1) t ->
  forall pf t1 t2,
    subtype t (typ_fun pf t1 t2) ->
    exists L, forall x, x `notin` L ->
    typing pf ((x,t1)::E) P (open e1 x) t2.
Proof.
  intros p E P e1 t T.
  remember (e_lam e1) as e'.
  generalize dependent e1.
  induction T; intros e' Eq pf' t1 t2 Sub; inversion Eq; subst.
  inversion Sub. subst. exists L. intros. eapply sub_narrowing. apply H5.
  eauto using typing_sub.
  apply IHT. reflexivity. apply (sub_transitivity T); assumption.
Qed.

Lemma typing_app_inv : forall p E P e1 e2 t2,
  typing p E P (e_app e1 e2) t2 ->
  exists p1, exists p2, exists p3, exists pf, exists t1, exists p',
  comb_cflow p1 p2 p3 /\
  comb_cflow p3 pf p' /\
  phi_sub p' p /\
  typing p1 E P e1 (typ_fun pf t1 t2) /\
  typing p2 E P e2 t1.
Proof.
  intros.
  remember (e_app e1 e2) as e.
  induction H; inversion Heqe; subst.
  Case "app".
    destruct H1 as [p3 [cf1 cf2]].
    exists p1. exists p2. exists p3. exists pf. exists T1. exists p.
    auto using phi_sub_refl.
  Case "sub".
    destruct (IHtyping H2) as [p1[p2[p3[pf[t1[p''[cf1[cf2[T1 T2]]]]]]]]].
    exists p1. exists p2. exists p3. exists pf. exists t1. exists p''.
    split. assumption. split. assumption.
    split. eauto using phi_sub_trans.
    destruct T2 as [T2a T2b].
    split; auto.
   eauto using typing_sub, sub_fun, sub_refl, phi_sub_refl.
Qed.

Lemma typing_ref_inv : forall p E P e t l,
  typing p E P (e_ref l e) t ->
  exists teff, exists t', subtype (typ_ref teff t') t /\
  typing p E P e t' /\ l `in` teff.
Proof.
  intros.
  remember (e_ref l e) as e'.
  induction H; inversion Heqe'; subst.
  Case "sub".
    destruct (IHtyping H2) as [teff [t' [sub [T1 I]]]].
    exists teff. exists t'. split.
      eapply (sub_transitivity T'); assumption.
      split. eauto using typing_sub, sub_refl. assumption.
  Case "ref".
    exists (singleton l). exists t.
    split. apply sub_refl.
    split. assumption. decideFSet.    
Qed.

Lemma typing_deref_inv : forall p E P e t,
  typing p E P (e_deref e) t ->
  exists p1, exists fst, exists teff, exists rst, exists t', exists p',
  subtype t' t /\
  phi_sub p' p /\
  typing p1 E P e (typ_ref teff t') /\
  comb_cflow p1 (make_phi fst teff rst) p'.
Proof.
  intros.
  remember (e_deref e) as e'.
  induction H; inversion Heqe'; subst.
  Case "sub".
    destruct (IHtyping H2) as [p1[fst[teff[rst[t'[p''[Sub [SP [T1 Cf]]]]]]]]].
    exists p1. exists fst. exists teff. exists rst. exists t'. exists p''.
    split. apply (sub_transitivity T'); assumption.
    split. apply (phi_sub_trans p'' p'); assumption.
    split; assumption.
  Case "deref".
    exists p1. exists fst. exists teff. exists rst. exists t. exists p.
    auto 6 using sub_refl, phi_sub_refl.
Qed.

Lemma typing_assign_inv : forall p E P e1 e2 t,
  typing p E P (e_assign e1 e2) t ->
  exists t', exists p', exists p1, exists p2, exists p3,
    exists fst, exists teff, exists rst,
      comb_cflow p1 p2 p3 /\
      comb_cflow p3 (make_phi fst teff rst) p' /\
      subtype t' t /\
      phi_sub p' p /\
      typing p1 E P e1 (typ_ref teff t') /\
      typing p2 E P e2 t'.
Proof.
  intros.
  remember (e_assign e1 e2) as e'.
  induction H; inversion Heqe'; subst.
  Case "sub".
    destruct (IHtyping H2)
      as [t'[p''[p1[p2[p3[fst[teff[rst[Cf1[Cf2[Sub[PS[T1 T2]]]]]]]]]]]]].
    exists t'. exists p''. exists p1. exists p2. exists p3.
    exists fst. exists teff. exists rst.
    split. assumption.
    split. assumption.
    split. apply sub_transitivity with T'; auto.
    split. apply (phi_sub_trans p'' p'); assumption.
    split; assumption.
  Case "assign".
    exists t. exists p. exists p1. exists p2. exists p'.
    exists fst. exists teff. exists rst.
    auto 6 using sub_refl, phi_sub_refl.
Qed.
    
Lemma typing_loc_inv : forall p E P r l t,
  typing p E P (e_loc r l) t ->
  ok E /\ ok P /\
  exists t',
    subtype (typ_ref (singleton l) t') t /\
    binds r t' P.
Proof.
  intros. remember (e_loc r l) as e'.
  induction H; inversion Heqe'; subst.
  Case "sub".
    destruct (IHtyping H2) as [okE[okP[t'[Sub B ]]]].
    repeat split;auto.
    exists t'. split. apply (sub_transitivity T'); assumption. assumption.
  Case "loc".
    split. assumption.
    split. assumption.
    exists t. split. apply sub_refl. assumption.
Qed.


Lemma typing_weak_P : forall p E P e t r t',
  r `notin` dom P ->
  typing p E P e t ->
  typing p E ((r,t')::P) e t.
Proof.
  intros p E P e t r t' NE T. induction T; subst; eauto.
  apply typing_loc; auto. simpl_env. apply binds_tail. assumption.
  simpl. destruct (ptr0 == r).
    subst. unfold not in NE. assert False. apply NE. eapply binds_In. apply H1.
    destruct H2.
    decideFSet.
Qed.

Lemma heap_push : forall r v t P H,
  heap_ok P H ->
  r `notin` dom P ->
  value v ->
  typing phi_0 nil P v t ->
  heap_ok ((r,t)::P) ((r,v)::H).
Proof.
  intros. unfold heap_ok in *. destruct H0 as [Pok [Eq ?]].
  split. eauto using ok_cons, env_ok.
  split. simpl. rewrite Eq. reflexivity.
  intros. destruct (r == r0).
    subst. exists t. exists v.
      split. assumption. 
      split. simpl_env. auto using binds_head, binds_singleton.
      split. simpl_env. auto using binds_head, binds_singleton.
      intros. apply typing_weak_P; auto.
    simpl in H4. assert (r0 `in` dom H). remember (dom H) as a. decideFSet.
    destruct (H0 r0 H5) as [t'[v'[V[bh[bp T]]]]].
    exists t'. exists v'. simpl_env.
    split. assumption.
    split. apply binds_tail. assumption. decideFSet.
    split. apply binds_tail. assumption. decideFSet.
    simpl. apply typing_weak_P. assumption. apply T.
Qed.

Lemma binds_unique : forall A, forall r (x x':A) h,
  binds r x h -> binds r x' h -> x = x'.
Proof.
  intros. unfold binds in *. induction h.
  rewrite H in H0. inversion H0. subst. reflexivity.
  destruct a as (r',x''). simpl in *. destruct (r == r').
    subst. inversion H; inversion H0; subst. assumption. auto.
Qed.

Ltac break_heap_ok H r :=
  match type of H with
    | heap_ok ?P ?h =>
        let okP := fresh "okP" in
        let domeq := fresh "domeq" in
        let H' := fresh "H" in
        let t := fresh "t" in
        let v := fresh "v" in
        let V := fresh "V" in
        let bh := fresh "bh" in
        let bp := fresh "bp" in
        destruct H as [okP [ domeq H' ]];
        destruct (H' r) as [t [ v [ V [ bh [ bp T]]]]];
        [ eauto 3 using binds_In; try rewrite domeq; eauto 3 using binds_In
        | clear H' ]
  end.

Tactic Notation "break_heap_ok" "in" hyp(H) "for" hyp(r) hyp(v') hyp(t') :=
  match type of H with
    | heap_ok ?P ?h =>
        let okP := fresh "okP" in
        let domeq := fresh "domeq" in
        let H' := fresh "H" in
        let t := fresh "t" in
        let v := fresh "v" in
        let V := fresh "V" in
        let bh := fresh "bh" in
        let bp := fresh "bp" in
        destruct H as [okP [ domeq H' ]];
        destruct (H' r) as [t [ v [ V [ bh [ bp T]]]]];
        [ eauto 3 using binds_In; try rewrite domeq; eauto 3 using binds_In
        | clear H';
          assert (v = v');
          [ eauto using binds_unique
            | assert (t = t'); [eauto using binds_unique | subst]
          ]
        ]
  end.
  

Lemma value_in_heap : forall P heap r v,
  heap_ok P heap -> binds r v heap -> value v.
Proof.
  intros. break_heap_ok H r.
  assert (v = v0). eauto using binds_unique.
  subst. assumption.
Qed.

Lemma typing_in_heap : forall P heap r v t,
  heap_ok P heap -> binds r v heap -> binds r t P ->
  typing phi_0 nil P v t.
Proof.
  intros. break_heap_ok in H for r v t. auto.
Qed.

Lemma heap_ok_binds : forall P h r,
  heap_ok P h -> ((exists v, binds r v h) <-> (exists t, binds r t P)).
Proof.
  intros. split; intros [v B]; break_heap_ok H r.
    exists t. assumption.
    exists v0. assumption.
Qed.

Lemma dom_after_replace: forall h r v v',
  binds r v h ->
  dom h = dom (replace h r v').
Proof.
  intros.
  induction h as [|(l,e)]. auto.
  simpl. destruct (l == r). subst. simpl. auto.
  simpl. rewrite IHh. reflexivity. unfold binds in H. simpl in H.
    destruct (r == l). destruct n. auto. unfold binds. auto.
Qed.

Lemma heap_ok_after_replace: forall P h r v v' t,
  heap_ok P h ->
  binds r v h ->
  binds r t P ->
  value v' ->
  typing phi_0 nil P v' t ->
  heap_ok P (replace h r v').
Proof.
  intros.
  destruct H as [Ok[Domeq HH]].
  unfold heap_ok. rewrite <- (dom_after_replace h r v v') in *; auto.
    split. assumption.
    split. assumption.
    intros. destruct (HH r0 H) as [t0[v0[V0[Bh0[Bp0 T0]]]]]. clear HH.
    exists t0.
    destruct(r == r0).
    Case "equal".
      subst r0. exists v'.
      split. assumption.
      split. apply binds_after_replace with v0. assumption.
      split. assumption.
      assert (t0 = t) as Eq. eauto using binds_unique. subst. assumption.
    Case "not equal".
      exists v0.
      split. assumption.
      split. apply binds_other_after_replace; auto.
      split. auto.
      assumption.
Qed.

Ltac standard_soundness_value v := 
  exists (nil:env); exists v; simpl in *; eauto 10 using empty_set_subset.

Ltac propfullset eq Eq1 Eq2 D :=
    assert eq as Eq1;
    [ apply subseteq; 
        [ apply subset_full_set
          | rewrite <- Eq2;
            eauto using forward_effect_subset ]
    | idtac ].

Ltac propfullset2 eq Eq1 Eq2 D :=
    assert eq as Eq1;
    [ apply subseteq; 
        [ apply subset_full_set
          | rewrite <- Eq2;
            eauto using backward_effect_subset ]
    | idtac ].

Theorem standard_effect_soundness :
  forall p P H H' e teff t R fst rst fst' rst',
    typing p nil (P) e t ->
    heap_ok P H ->
    evaluation fst rst H e teff fst' rst' H' R ->
    fst = full_set ->
    rst = full_set ->
    fst' = full_set ->
    rst' = full_set ->
    exists P', exists v,
      R = exp_result v /\
      value v /\
      typing phi_0 nil (P' ++ P) v t /\
      heap_ok (P' ++ P) H' /\
      lset_subset teff (phi_fld_eff p)
.
Proof.
  intros p P H H' e teff t R fst rst fst' rst' T Hok D.
  intros Eqfst Eqrst Eqfst' Eqrst'.
  generalize dependent p.
  generalize dependent t.
  generalize dependent P.
  induction D; intros P Hok t p T.
  Case "Id".
    exists (nil:env). exists v.
      eauto 6 using generalize_value_phi, empty_set_subset.
  Case "app".
    destruct (typing_app_inv p nil P e1 e2 t T)
      as [p1 [p2 [p3 [pf [t1 [p' [Cf1 [Cf2 [Sub [T1 T2]]]]]]]]]].
    propfullset (fst1 = full_set) Eqfst1 Eqfst D1.
    propfullset (fst2 = full_set) Eqfst2 Eqfst1 D2.
    propfullset2 (rst2 = full_set) Eqrst2 Eqrst' D3.
    propfullset2 (rst1 = full_set) Eqrst1 Eqrst2 D2.
    destruct (IHD1 Eqfst Eqrst Eqfst1 Eqrst1 P Hok (typ_fun pf t1 t) p1 T1) as
      [ P1 [ vlam [ Nerr1 [ Vlam [Tf [ H1ok Sub1 ] ]] ]]].
    inversion Nerr1. subst vlam.
      assert (typing p2 nil (P1 ++ P) e2 t1).
      apply (typing_weakening p2 nil nil P P1 e2 t1 T2).
        simpl. auto.
        apply (env_ok heap1 (P1 ++ P) H1ok).
      destruct (IHD2 Eqfst1 Eqrst1 Eqfst2 Eqrst2 (P1 ++ P) H1ok t1 p2 H) 
        as [P2 [v2' [Nerr2 [V2 [Tv2 [H2ok Sub2]]]]]].
      inversion Nerr2. subst v2'.
      assert (typing phi_0 nil (P2 ++ P1 ++ P) (e_lam e) (typ_fun pf t1 t)) as Tf'.
      apply (typing_weakening phi_0 nil nil (P1 ++ P) P2 (e_lam e)
                 (typ_fun pf t1 t) Tf); simpl.
        apply ok_nil.
        apply env_ok with heap2. assumption.
        destruct (typing_lam_inv _ _ _ _ _ Tf' pf t1 t 
                (sub_refl (typ_fun pf t1 t)))
          as [L Tpf].
      assert (typing pf nil (P2 ++ P1 ++ P) (open e v2) t).
        pick fresh y.
        rewrite (subst_intro y).
        apply typing_subst with t1. apply Tpf. decideFSet. assumption.
        decideFSet.
        apply typing_regular_lc with (p := phi_0) (E := nil:env) (P := P2 ++ P1 ++ P) (T := t1). apply Tv2.
      destruct (IHD3 Eqfst2 Eqrst2 Eqfst' Eqrst' (P2 ++ P1 ++ P) H2ok t pf H0)
        as [P3 [v' [Nerr3 [V [Tv [H3ok Sub3]]]]]].
      inversion Nerr3. subst v'.
      exists (P3 ++ P2 ++ P1). exists v. simpl_env.
      split. reflexivity.
      split. assumption.
      split. apply Tv.
      split. apply H3ok.
      inversion Cf1. inversion Cf2.
      rewrite <- H1 in *.
      rewrite <- H2 in *.
      rewrite <- H3 in *.
      rewrite <- H4 in *.
      rewrite <- H5 in *.
      rewrite <- H6 in *.
      inversion H4.
      inversion Sub. rewrite <- H14 in *. rewrite H9 in *. simpl in *.
      rewrite <- set_union_assoc in H15.
      apply subset_transitive with (s2 := (teff0 `union` teff4 `union` teff6)).
      apply subset_add. assumption.
        apply subset_add; assumption.
        assumption.
  Case "ref".
    destruct (typing_ref_inv p nil P e t l T) as [eff [t' [Sub [T' I]]]].
    destruct (IHD Eqfst Eqrst Eqfst' Eqrst' P Hok t' p T') as [P' [v' [Nerr [V [p0 [H'ok Sub1]]]]]].
    inversion Nerr. subst v'.
    exists ((r,t'):: P'). exists (e_loc r l).
    split. reflexivity.
    split. apply loc_value.
    split. intros.
    apply typing_sub with (p' := phi_0) (T' := typ_ref (singleton l) t').
    apply typing_loc.
      apply ok_nil. simpl_env. apply ok_push. apply env_ok with heap'. assumption.
      destruct H'ok as [? [Dom ?]].
      rewrite <- Dom. auto. simpl_env. auto using binds_head, binds_singleton.
      apply (sub_transitivity (typ_ref eff t')).
      apply sub_ref.
      apply sub_refl.
      apply sub_refl.
      unfold lset_subset. intros.
      (* coq (proof general) breaks after this, step-back does not work. *)
      decideFSet.
      assumption.
      apply phi_sub_refl.
    split. simpl. apply heap_push; auto.
      destruct H'ok as [_ [Dom _]]. rewrite <- Dom. assumption.
      assumption.
  Case "deref".
    destruct (typing_deref_inv p nil P e t T)
      as [p1[fst1[teff1[rst1[t'[p'[Sub[PS[T' Cf]]]]]]]]].
    propfullset (fst' = full_set) Eqfst'' Eqfst D.
    assert (rst' `union` singleton label0 = full_set) as Eqrst''.
      subst rst'. apply union_full_set.
    destruct (IHD Eqfst Eqrst Eqfst'' Eqrst'' P Hok (typ_ref teff1 t') p1 T')
      as [P1[v1[Nerr1[V1[Tf[H1ok Sub1]]]]]]. clear IHD.
    inversion Nerr1. subst v1. clear Nerr1.
    destruct (typing_loc_inv phi_0 nil (P1 ++ P) ptr0 label0 (typ_ref teff1 t') Tf)
      as [_[_ [t'' [Sub2 B]]]]. clear Tf.
    exists P1. exists v.
    split. reflexivity.
    split. eauto using value_in_heap.
    split. apply typing_sub with (p' := phi_0) (T' := t'').
      apply typing_in_heap with (heap := heap') (r := ptr0). assumption.
      assumption.
      apply B.
      inversion Sub2. apply (sub_transitivity t'); assumption. apply phi_sub_refl.
    split. assumption.
    inversion Cf. subst p'. subst p1. clear Cf.
    inversion Sub2. subst teff'. subst teff3. clear Sub2.
    inversion PS. subst p. clear PS. simpl in *.
    apply subset_transitive with (s2 := (teff0 `union` teff1)).
    apply subset_add. assumption. assumption. assumption.
  Case "assign".
    destruct (typing_assign_inv p nil P e1 e2 t T)
      as [t'[p'[p1[p2[p3[fst'[teff[rst'[Cf1[Cf2[Sub1[PS[T1 T2]]]]]]]]]]]]].
    subst.
    propagate_full_set fst1.
    propagate_full_set fst2.
    propagate_full_set rst1.
    assert (full_set = full_set) as EQ. reflexivity.
    destruct (IHD1 EQ EQ EQ EQ P Hok (typ_ref teff t') p1 T1)
      as [ P1 [ r [ Nerr1 [Vr [Tr[H1ok Sub2]]]]]]. clear IHD1.
    inversion Nerr1. subst r. clear Nerr1.
    assert (typing p2 nil (P1 ++ P) e2 t') as T2'.
      apply (typing_weakening p2 nil nil P P1 e2 t' T2).
        simpl. auto.
        apply env_ok with heap1. assumption.
        clear T2.
    assert (full_set `union` (singleton label0) = full_set) as EQ1.
      apply union_full_set.
    destruct (IHD2 EQ EQ EQ EQ1 (P1 ++ P) H1ok t' p2 T2')
      as [P2[v2[Nerr2[Vv2[Tv2[H2ok Sub3]]]]]]. clear IHD2.
     inversion Nerr2. subst v2. clear Nerr2 EQ.
    destruct (typing_loc_inv phi_0 nil (P1 ++ P) ptr0 label0 (typ_ref teff t') Tr)
      as [_[_ [t'' [Sub4 B]]]]. clear Tr.
    exists (P2 ++ P1). exists v. simpl_env.
    split. reflexivity.
    split. assumption.
    assert (t'' = t') as Eq.
      inversion Sub4. subst. apply subeq; assumption. subst t''.
    split. apply typing_sub with (p' := phi_0) (T' := t'); auto using phi_sub_refl.
    split. apply heap_ok_after_replace with v' t'; auto.
    apply binds_concat_ok. assumption.
    destruct H2ok as [P2ok _]. assumption.
    inversion Sub4. subst.
    clear D1 D2 Vv2 Tv2 Sub4 B T1 Sub1 T t H Hok H1ok T2' H2ok P EQ1.
    inversion Cf1. subst p1 p2 p3.
    inversion Cf2. subst p' teff5 rst0 teff4 fst0 rst2. clear Cf1 Cf2.
    inversion PS. subst p rst1 fst0 teff4 fst'. clear PS. simpl in *.
    apply subset_transitive with ((teff0 `union` teff3) `union` teff);
      auto using subset_add.
  Case "call-w".
    destruct (typing_app_inv p nil P e1 e2 t T)
      as [p1 [p2 [p3 [pf [t1 [p' [Cf1 [Cf2 [Sub [T1 T2]]]]]]]]]].
    destruct (IHD Eqfst Eqrst Eqfst' Eqrst' P Hok (typ_fun pf t1 t) p1 T1) as
      [ P1 [ vlam [ Nerr1 [ Vlam [Tf [ H1ok Sub1 ] ]] ]]]. clear IHD.
    inversion Nerr1. subst v. clear Nerr1.
    inversion Vlam; subst vlam.
    Subcase "nat".
    destruct (typing_nat_inv phi_0 n nil (P1++P) (typ_fun pf t1 t) Tf) as [_[_ S]].
    inversion S.
    Subcase "lam".
    subst. destruct (H e). reflexivity.
    Subcase "loc".
    destruct (typing_loc_inv phi_0 nil (P1++P) ptr0 label0 (typ_fun pf t1 t) Tf)
      as [_[_[t'[S _]]]].
    inversion S.
  Case "deref_h_w".
    destruct (typing_deref_inv p nil P e t T)
      as [p1[a[ee[o[t'[p'[S[PS[T' C]]]]]]]]].
    destruct (IHD Eqfst Eqrst Eqfst' Eqrst' P Hok (typ_ref ee t') p1 T') as [P1[v[Eq[V[Tf[H1ok Sub]]]]]].
    inversion Eq. subst v. clear Eq. clear IHD.
    destruct (typing_loc_inv phi_0 nil (P1 ++ P) ptr0 label0 (typ_ref ee t') Tf)
      as [_[_[t''[S' B]]]].
    destruct (heap_ok_binds (P1 ++ P) heap' ptr0 H1ok). clear H0.
    destruct H1. exists t''. assumption.
    assert False. apply H. apply binds_In with x. assumption. inversion H1.
  Case "deref_l_w".
    destruct H0. rewrite Eqfst'. apply everything_in_full_set.
  Case "assign_h_w".
    destruct (typing_assign_inv p nil P e1 e2 t T)
      as [t'[p'[p1[p2[p3[fst'[teff[rst'[Cf1[Cf2[Sub1[PS[T1 T2]]]]]]]]]]]]].
    subst.
    propagate_full_set fst1.
    propagate_full_set rst1.
    assert (full_set = full_set) as EQ. reflexivity.
    destruct (IHD1 EQ EQ EQ EQ P Hok (typ_ref teff t') p1 T1)
      as [ P1 [ r [ Nerr1 [Vr [Tr[H1ok Sub2]]]]]]. clear IHD1.
    inversion Nerr1. subst r. clear Nerr1.
    assert (typing p2 nil (P1 ++ P) e2 t') as T2'.
      apply (typing_weakening p2 nil nil P P1 e2 t' T2).
        simpl. auto.
        apply env_ok with heap1. assumption.
        clear T2.
    assert (full_set `union` (singleton label0) = full_set) as EQ1.
      apply union_full_set.
    destruct (IHD2 EQ EQ EQ EQ (P1 ++ P) H1ok t' p2 T2')
      as [P2[v2[Nerr2[Vv2[Tv2[H2ok Sub3]]]]]]. clear IHD2.
     inversion Nerr2. subst v2. clear Nerr2 EQ.
    destruct (typing_loc_inv phi_0 nil (P1 ++ P) ptr0 label0 (typ_ref teff t') Tr)
      as [_[_ [t'' [Sub4 B]]]]. clear Tr.
    destruct H2ok as [P2ok[Domeq _ ]].
    assert False. apply H. rewrite Domeq. apply binds_In with t''.
    apply binds_concat_ok. assumption. assumption. destruct H0.
  Case "assign_l_w".
    destruct H0. subst rst2. apply everything_in_full_set.
Qed.