Require Export Lair_Util.
  
(** An environment G,A is well-formed, denoted [(wf_env G A)], if each
    atom is bound at most at once and if each binding is to a
    well-formed type. *)

Inductive wf_env : gamma -> atoms -> Prop :=
| wf_env_empty :
  wf_env nil {}

| wf_env_typ : 
  forall (G : gamma) (A : atoms) (X : atom) (T : typ),
    wf_env G A ->
    X `notin` dom G ->
    wf_typ T G ->
    wf_env ([(X, binding_typ T)] ++ G) A
    
| wf_env_typ_A : 
  forall (G : gamma) (A : atoms) (X : atom) (T : typ),
    wf_env G A ->
    X `notin` dom G ->
    wf_typ T G -> 
    wf_env ([(X, binding_typ T)] ++ G) ((singleton X) `union` A)

| wf_env_kind : 
  forall (G : gamma) (A : atoms) (X : atom) (K : kind),
    wf_env G A ->
    X `notin` dom G ->
    wf_env ([(X, binding_kind K)] ++ G) A.

(* ********************************************************************** *)
(* wf_env generalizes the ok predicate of Metatheory;                     *)
(* Useful for working with lemmas in the Environment library              *)
(* ********************************************************************** *)

Lemma wf_is_ok : forall g A, 
  wf_env g A -> ok g.
Admitted.

Lemma binds_exists_singleton : forall x (b:binding) G, 
  binds x b G -> exists G1, exists G2, G = G1 ++ [(x, b)] ++ G2.
Admitted.

(* ********************************************************************** *)
(* Lemmas about splitting assumptions *)
(* ********************************************************************** *)

Lemma split_nil_nil : forall A1 A2, splitA A1 A2 {} -> A1 = {} /\ A2 = {}.
Admitted. 

Lemma split_nil_A_1 : forall A, splitA {} A A.
Admitted. 

Lemma split_nil_A_2 : forall A, splitA A {} A.
Admitted. 

Lemma split_nil_eq_1 : forall A B, splitA {} A B -> A = B.
Admitted.

Lemma split_nil_eq_2 : forall A B, splitA A {} B -> A = B.
Admitted.

Lemma split_commutes : forall A B C, splitA A B C -> splitA B A C.
Admitted. 

Lemma split_trans_1 : forall A1 A2 A2_1 A2_2 A, 
  splitA A1 A2 A -> splitA A2_1 A2_2 A2 -> exists B, splitA A1 A2_1 B.
Admitted.

Lemma split_trans_2 : forall A1 A2 A2_1 A2_2 A, 
  splitA A1 A2 A -> splitA A2_1 A2_2 A2 -> exists B, splitA A1 A2_2 B.
Admitted.

Lemma split_trans_3 : forall A1 A2 A3 A4 A5, 
  splitA A1 A2 A3 -> splitA A4 A3 A5 -> exists A, splitA A4 A1 A.
Admitted.

Lemma split_trans_4 : forall A1 A2 A3 A4 A5, 
  splitA A1 A2 A3 -> splitA A4 A3 A5 -> exists A, splitA A4 A2 A.
Admitted.

Lemma split_trans_5 : forall A1 A2 A12 A2_1 A2_2 A1__A2_1 ,
  splitA A1 A2 A12 -> splitA A2_1 A2_2 A2 -> splitA A1 A2_1 A1__A2_1 -> splitA A1__A2_1 A2_2 A12.
Admitted.

Lemma split_trans_6 : forall A1 A2 A12 A2_1 A2_2 A1__A2_2 ,
  splitA A1 A2 A12 -> splitA A2_1 A2_2 A2 -> splitA A1 A2_2 A1__A2_2 -> splitA A2_1 A1__A2_2 A12.
Admitted.

Lemma split_trans_7 : forall A1 A2 A Amx A1mx x, 
  splitA A1 A2 A -> A = singleton x `union` Amx -> A1 = singleton x `union` A1mx -> 
  x `notin` Amx -> x `notin` A1mx -> splitA A1mx A2 Amx.
Admitted.

Lemma split_trans_8 : forall A1mx A2mx Amx x, 
  splitA A1mx A2mx Amx -> x `notin` Amx -> 
  splitA (singleton x `union` A1mx) A2mx (singleton x `union` Amx).
Admitted.

Lemma split_trans_9 : forall a a0 fx_e fx_e' fx_newe fx0,
  splitA (singleton a) fx_e' fx_e -> 
  splitA (singleton a0) fx_e fx_newe -> 
  splitA (singleton a) fx0 fx_newe -> 
  splitA (singleton a0) fx_e' fx0.
Admitted.

Lemma split_trans_10 : forall a a0 fx_e fx_newe fx0, 
  a <> a0 -> 
  splitA (singleton a0) fx_e fx_newe -> 
  splitA (singleton a) fx0 fx_newe -> 
  exists fx_e', splitA (singleton a) fx_e' fx_e.
Admitted.

Lemma split_extension_R : forall A B C x, 
  splitA A B C -> x `notin` C -> splitA A (B `union` singleton x) (C `union` singleton x).
Admitted.

Lemma splitA_splits : forall A B C, 
  splitA A B C -> (forall x, (x `in` A -> x `notin` B) /\ (x `in` B -> x `notin` A)).
Proof.
intros.
         inversion H.
         split. 
           intro x_in_A. unfold not. intro x_in_B. 
           pose proof (AtomSet.F.inter_3 x_in_A x_in_B) as for_contra. rewrite H0 in for_contra; fsetdec.
           intro x_in_B. unfold not. intro x_in_A. 
           pose proof (AtomSet.F.inter_3 x_in_A x_in_B) as for_contra. rewrite H0 in for_contra; fsetdec.
Qed.

  

(* ********************************************************************** *)
(* Lemmas about the initial typing environment that contains only names *)
(* ********************************************************************** *)

Lemma only_names_split1 : forall (G:gamma) (A1:atoms) (A2:atoms) (A:atoms), 
  only_names G A -> splitA A1 A2 A -> only_names G A1.
Admitted.

Lemma only_names_split2 : forall (G:gamma) (A1:atoms) (A2:atoms) (A:atoms), 
  only_names G A -> splitA A1 A2 A -> only_names G A2.
Admitted.

Lemma binds_no_vars : forall (G:gamma) (A:atoms) (x:atom) (t:typ), 
  only_names G A -> not (binds x (binding_typ t) G).
  intros G A x t ON. induction ON.
  Case "empty".
  unfold not. intros. inversion H.
  Case "extra".
  unfold not. intros binds_x_XG.
  destruct (binds_concat_inv binding x (binding_typ t) G ([(X, binding_kind kind_N)]) binds_x_XG) as [ bindsG | bindsHD ].
  destruct bindsG as [xNotinHd binds_x_F]. auto.
  inversion bindsHD.
  destruct (x==X); discriminate.
  Case "cons".
  unfold not. intros binds_x_XG.
  destruct (binds_concat_inv binding x (binding_typ t) G ([(X, binding_kind kind_N)]) binds_x_XG) as [ bindsG | bindsHD ].
  destruct bindsG as [xNotinHd binds_x_F]. auto.
  inversion bindsHD.
  destruct (x == X); discriminate.
Qed.

Lemma binds_no_vars_sub : forall G1 G2 x t G1_1 G1_2, 
   G1_2 ++ G1_1 = G2 ++ ([(x, binding_typ t)]) ++ G1 ->
   only_names G1_1 {} ->  x `notin` dom G1_1.
Admitted.

Lemma only_names_prefix : forall G1 G2 x t G1_1 G1_2, 
   G1_2 ++ G1_1 = G2 ++ ([(x, binding_typ t)]) ++ G1 ->
   only_names G1_1 {} ->  
   exists G1_1', G1 = G1_1' ++ G1_1.
Admitted.


(* ********************************************************************** *)
(* Lemmas about well-formed environments *)
(* ********************************************************************** *)

Lemma wf_split1 : forall (G:gamma) (A1:atoms) (A2:atoms) (A:atoms), 
  wf_env G A -> splitA A1 A2 A -> wf_env G A1.
Admitted.

Lemma wf_split2 : forall (G:gamma) (A1:atoms) (A2:atoms) (A:atoms), 
  wf_env G A -> splitA A1 A2 A -> wf_env G A2.
Admitted.

Lemma wf_cat : forall (G1 : gamma) (G2 : gamma) A A1 A2, 
  wf_env G1 A -> splitA A A1 A2 ->  wf_env (G2 ++ G1) A2 ->
  wf_env (G2 ++ G1) A.
Admitted.

Lemma wf_env_aqual : forall G A x t k, 
  wf_env G A -> binds x (binding_typ t) G -> wf_env G (aqual x k A).
Admitted.

Lemma wf_env_tail : forall G1 G2 A, 
  wf_env (G2 ++ G1) A -> wf_env G1 {}.
Admitted.

Module LAIR_Environment (Import X : LAIR_PARAMS).
  Lemma wf_env_extension : forall G A y k t, 
   wf_env G A -> y `notin` dom G -> 
   kinding G t k ->  wf_env  ([(y, binding_typ t)]++G) (aqual y k A).
  Admitted.
End LAIR_Environment.

Lemma wf_env_G_A_containment : forall G A, 
  wf_env G A -> AtomSet.F.Subset A (dom G).
intros.
generalize dependent A.
induction G.
Case "Gnil".
  intros.
  inversion H. simpl. fsetdec.
Case "Gcons".
  intros.
  inversion H.
  SCase "wf_env_typ".
    simpl.
    pose proof (IHG A H2) as IH. fsetdec.
  SCase "wf_env_typ_A".
    simpl.
    pose proof (IHG A0 H2) as IH. fsetdec.
  SCase "wf_env_kind".
    simpl.
    pose proof (IHG A H2) as IH. fsetdec.
Qed.

Lemma wf_env_containment : forall G A A', 
  wf_env G A -> AtomSet.F.Subset A' A -> wf_env G A'.
Admitted.

Lemma wf_exp_weakening : forall G1 G2 v, 
  wf_exp v G1 -> wf_exp v (G2 ++ G1).
Admitted.

Lemma wf_typ_weakening : forall G1 G2 t, 
  wf_typ t G1 -> wf_typ t (G2 ++ G1).
Admitted.

Lemma binds_wf : forall G A x t, 
  wf_env G A -> (binds x (binding_typ t) G) -> wf_typ t G.
intros.
generalize dependent A.
induction G.
Case "G_nil".
inversion H0.
intros.
Case "G_cons".
inversion H.
SCase "wf_env_typ".
subst.
rewrite_env ([(X, binding_typ T)] ++ G) in H0.
destruct (binds_concat_inv binding x (binding_typ t) G _ H0) as [ [xNotX0 G_binds_x] | x_is_X0].
(* G_binds_x *)
pose proof (IHG G_binds_x A H3) as wf_t_G.
rewrite_env ([(X, binding_typ T)] ++ G).
eapply wf_typ_weakening; auto.
(* x_is_X0 *)
inversion x_is_X0.
destruct (x == X).
  (* subcase x=X *)
  assert (t = T).
  injection H2; auto.
  subst.
  inversion H.
     (* SubCase : wf_env_typ *)
     rewrite_env ([(X, binding_typ T)] ++ G).
     eapply wf_typ_weakening. assumption.
     rewrite_env ([(X, binding_typ T)] ++ G).
     eapply wf_typ_weakening. assumption.
  (* subcase x <> X *)
  discriminate.
SCase "wf_env_typ_A".
subst.
rewrite_env ([(X, binding_typ T)] ++ G) in H0.
destruct (binds_concat_inv binding x (binding_typ t) G _ H0) as [ [xNotX0 G_binds_x] | x_is_X0].
(* G_binds_x *)
pose proof (IHG G_binds_x A0 H3) as wf_t_G.
rewrite_env ([(X, binding_typ T)] ++ G).
eapply wf_typ_weakening; auto.
(* x_is_X0 *)
inversion x_is_X0.
destruct (x == X).
  (* subcase x=X *)
  assert (t = T).
  injection H2; auto.
  subst.
  inversion H.
     (* SubCase : wf_env_typ *)
     rewrite_env ([(X, binding_typ T)] ++ G).
     eapply wf_typ_weakening. assumption.
     rewrite_env ([(X, binding_typ T)] ++ G).
     eapply wf_typ_weakening. assumption.
  (* subcase x <> X *)
  discriminate.
SCase "wf_env_kind".
subst.
rewrite_env ([(X, binding_kind K)] ++ G) in H0.
destruct (binds_concat_inv binding x (binding_typ t) G _ H0) as [ [xNotX G_binds_x] | x_is_X].
pose proof (IHG G_binds_x A H3) as wf_t_G.
rewrite_env ([(X, binding_kind K)] ++ G).
eapply wf_typ_weakening; auto.
inversion x_is_X. destruct (x == X); discriminate.
Qed.


Lemma binds_disjoint_1 : forall G1 x t G2 A, 
  wf_env (G2 ++ ([(x, binding_typ t)]) ++ G1) A -> x `notin` dom G1.
Admitted.

Lemma binds_disjoint_2 : forall G1 x t G2 A, 
  wf_env (G2 ++ ([(x, binding_typ t)]) ++ G1) A -> x `notin` dom G2.
Admitted.


(* ********************************************************************** *)
(* Lemmas about substitution over environments *)
(* ********************************************************************** *)


Lemma subst_typ_gamma_dom : forall G a t, 
  dom G = dom (subst_typ_gamma a t G).
Admitted.

Lemma subst_gamma_dom : forall G x v, 
  dom G = dom (subst_gamma x v G).
Admitted.

Lemma subst_wf_typ : forall G1 G2 t x tx v A,
  wf_env (G2 ++ [(x, binding_typ tx)] ++ G1) A -> 
  wf_typ t (G2 ++ [(x, binding_typ tx)] ++ G1) ->
  wf_exp v G1 -> wf_typ (subst_typ x v t) ((subst_gamma x v G2) ++ G1).
Admitted.

Lemma subst_gamma_kind_invariant : forall a k G x v, 
  binds a (binding_kind k) G -> binds a (binding_kind k) (subst_gamma x v G).
Admitted.

Lemma subst_typ_gamma_invariant : forall G a b A,
  wf_env G A -> a `notin` dom G -> (subst_typ_gamma a b G = G).
Admitted.

Lemma subst_gamma_wf : forall G1 G2 A x t_x v k, 
  wf_env (G2 ++ [(x, binding_typ t_x)] ++ G1) (aqual x k A) -> 
  wf_exp v G1 -> (x `notin` A) ->
  wf_env ((subst_gamma x v G2) ++ G1) A.
  intros G1 G2 A x t_x v k wf_G2_x_G1_Ax wf_v x_notin_A.
  generalize dependent A.
  induction G2.
  Case "G2nil".
  intros.
  simpl.
  simpl_env in wf_G2_x_G1_Ax.
  destruct (aqual_cases A x k) as [Ax_is_A | Ax ].
     SCase "Ax_is_A".
     rewrite Ax_is_A in wf_G2_x_G1_Ax.
     inversion wf_G2_x_G1_Ax. 
       (* case wf_env_typ *) assumption.
       (* case wf_env_typ_A *) rewrite <- H3 in x_notin_A. fsetdec.
     SCase "Ax".
     rewrite Ax in wf_G2_x_G1_Ax.
     inversion wf_G2_x_G1_Ax.
     SSCase "wf_env_typ".
       pose proof (wf_env_G_A_containment G1 (singleton x `union` A) H2) as for_contra. 
       destruct H4. unfold AtomSet.F.Subset in for_contra. 
       eapply for_contra. fsetdec.
     SSCase "wf_env_typ_A".
       pose proof (wf_env_G_A_containment G1 A0 H3) as A0_sub_G1.
       assert (x `notin` A0) as x_notin_A0.
       unfold AtomSet.F.Subset in A0_sub_G1.
       unfold not. intro. 
       pose proof (A0_sub_G1 x H6) as for_contra. contradiction.
       assert (A0 = A) as A0_eq_A. eapply union_identity. apply H2. auto. auto.
       subst A0. auto.
  Case "G2cons".
  intros.
  simpl.
  inversion wf_G2_x_G1_Ax.
       SCase "wf_env_typ".
       pose proof (IHG2 A H1 x_notin_A) as wf_TL.
       rewrite_env (([(X, binding_typ (subst_typ x v T))]) ++ (subst_gamma x v G2 ++ G1)).
       pose proof (dom_concat _ G1 G2) as domEq.
       pose proof (dom_concat _ G1 (subst_gamma x v G2)) as domSubEq.
       pose proof (subst_gamma_dom G2 x v) as domGSubEq.
       eapply wf_env_typ. assumption. rewrite domSubEq. 
       rewrite <- domGSubEq.  simpl_env in H2. fsetdec.
       pose proof (wf_exp_weakening G1 G2 v wf_v) as wf_v_G1G2.
       assert (dom (G2 ++ G1) = dom (subst_gamma x v G2 ++ G1)).
       rewrite domSubEq. rewrite <- domGSubEq. auto.
       simpl_env in H1. simpl_env in H4.
       eapply subst_wf_typ.  apply H1. apply H4. apply wf_v.
       SCase "wf_env_typ_A".
       rewrite_env (([(X, binding_typ (subst_typ x v T))]) ++ (subst_gamma x v G2 ++ G1)).
       assert (X `notin` A0) as X_notin_A0.
         simpl_env in H2.
         assert (AtomSet.F.Subset A0 (dom (G2 ++ ([(x, binding_typ t_x)]) ++ G1))) as A0_subset.
           apply (wf_env_G_A_containment _ _ H2).
        simpl_env in A0_subset. simpl_env in H3. 
        unfold not. intro.
        unfold AtomSet.F.Subset in A0_subset.
        pose proof (A0_subset X H5) as for_contra. contradiction.
       destruct (aqual_union_identity_1 X A0 x k A H1 X_notin_A0 x_notin_A) as [A1 [A0_eq_aqxA1 x_notin_A1]].
       subst A0.
       assert (x <> X) as X_neq_x.
         unfold not; intro.
         subst x. 
         simpl_env in H3. fsetdec.
       destruct (aqual_union_identity_2 X x k  _ A H1 X_neq_x) as [A2 [A_eq_XA2 X_notin_A2]].
       assert (X `notin` A1) as X_notin_A1.
         unfold not; intro.
         pose proof (aqual_identity_0 X x k A1 H5) as for_contra. contradiction.
       assert (A1 = A2) as A1_eq_A2.
         eapply aqual_identity_3. apply X_notin_A1. assumption. apply x_notin_A1. apply x_notin_A. apply H1. apply (sym_equal A_eq_XA2).
       subst A2.
       pose proof (IHG2 A1 H2 x_notin_A1) as wf_TL.
       pose proof (dom_concat _ G1 G2) as domEq.
       pose proof (dom_concat _ G1 (subst_gamma x v G2)) as domSubEq.
       pose proof (subst_gamma_dom G2 x v) as domGSubEq.
       subst A.
       eapply wf_env_typ_A. assumption.
       rewrite domSubEq. rewrite <- domGSubEq. 
         simpl_env in H3. fsetdec.
       eapply subst_wf_typ. 
         simpl_env in H2. apply H2.
         simpl_env in H4. apply H4.
         apply wf_v.
       SCase "wf_env_kind".
       pose proof (IHG2 A H1 x_notin_A).
       rewrite_env ([(X, binding_kind K)] ++ subst_gamma x v G2 ++ G1).
       eapply wf_env_kind. assumption.
       pose proof (dom_concat _ G1 G2) as domEq.
       pose proof (dom_concat _ G1 (subst_gamma x v G2)) as domSubEq.
       pose proof (subst_gamma_dom G2 x v) as domGSubEq.
       rewrite domSubEq. rewrite <- domGSubEq. 
       simpl_env in H3. fsetdec.
Qed.

Lemma Gamma_subst_binding_typ : forall G x y t v, 
  (binds x (binding_typ t) G) -> (y <> x) -> (binds x (binding_typ (subst_typ y v t)) (subst_gamma y v G)).
Admitted.

Lemma Gamma_subst_typ_binding_kind : forall G x k y t, 
  (binds x (binding_kind k) G) -> (y <> x) -> (binds x (binding_kind k) (subst_typ_gamma y t G)).
Admitted.
