Require Export Lair_Definitions.
Require Import Lair_IndHyp.


(* ********************************************************************** *)
(* Utilities about sets *)
(* ********************************************************************** *)

Ltac setRewrite :=
  match goal with
  | H1 : ?A = ?B, H2 : ?a `in` ?S |- _ =>
    let H := fresh in
    try (((assert (a `in` A) as H by fsetdec) ; rewrite H1 in H)
      || ((assert (a `in` B) as H by fsetdec) ; rewrite <- H1 in H))
  end.

Ltac settac :=
  repeat progress (
    repeat intro ;
    try split ;
    try match goal with
        | |- ?x = ?y => apply eq_if_Equal
        end
  ) ; try setRewrite ; try fsetdec.

Lemma union_identity : forall x A B, 
  (singleton x `union` A) = (singleton x `union` B) -> 
  x `notin` A -> x `notin` B -> A = B.
Proof.
  settac.
Qed.

Lemma split_singleton : forall x A, 
  x `in` A -> exists A0, x `notin` A0 /\ A = (singleton x) `union` A0.
Proof.
  intros. exists (AtomSet.F.remove x A). split ; fsetdec.
Qed.

Lemma inter_distributes : forall A B C, 
  A `inter` B `union` C = {} -> (A `inter` C = {}) /\ (A `inter` B = {}).
Proof.
  settac.
Qed.

Lemma inter_symmetric : forall A B C, 
  A `inter` B  = C -> B `inter` A = C.
Proof.
  intros. fsetdec.
Qed.

Lemma inter_distributes_inv : forall A B C, 
  (A `inter` B = {}) -> (A `inter` C = {}) -> (A `inter` (B `union` C) = {}).
Proof.
  settac.
  assert (A `inter` B `union` C = (A `inter` B) `union` A `inter` C) by fsetdec.
  rewrite H2 in H1. rewrite H in H1. rewrite H0 in H1. fsetdec.
Qed.

Lemma empty_union : forall A B, 
  AtomSet.F.is_empty (A `union` B) = true -> (A = {}) /\ (B = {}).
Proof.
  settac ; pose proof (AtomSet.F.is_empty_2 H) ; fsetdec.
Qed.

(* I think that this one might not be constructively provable *)
Lemma non_empty_1 : forall A, 
 AtomSet.F.is_empty A = false -> exists x, x `in` A.
Admitted.

Lemma is_empty_total : forall A, AtomSet.F.is_empty A = true \/ AtomSet.F.is_empty A = false.
Proof.
  intros. destruct AtomSet.F.is_empty ; auto.
Qed.

Lemma empty_unique : forall s, AtomSet.F.Empty s -> s = {}.
Proof.
  intros. fsetdec.
Qed.

Lemma non_empty_2 : forall A, 
  A <> {} -> exists x, x `in` A.
Proof.
  intros. assert (AtomSet.F.is_empty A = false). pose proof is_empty_total A. destruct H0.
    pose proof AtomSet.F.is_empty_2 H0. pose proof empty_unique A H1. contradiction.
    assumption.
  apply non_empty_1. assumption.
Qed.

Lemma non_empty_3 : forall A x, 
  x `in` A -> AtomSet.F.is_empty A = false.
Proof.
  intros. pose proof is_empty_total A. destruct H0.
    pose proof AtomSet.F.is_empty_2 H0. fsetdec.
    assumption.
Qed.

Lemma demorgan_1 : forall X Y Z, 
  X `inter` (Y `union` Z) = (X `inter` Y) `union` (X `inter` Z).
Proof.
  intros. fsetdec.
Qed.

Lemma union_commutes : forall A B, 
  A `union` B = B `union` A.
Proof.
  intros; fsetdec.
Qed.

Lemma intersection_commutes : forall A B, A `inter` B = B `inter` A.
Proof.
  intros; fsetdec.
Qed.

Lemma union_associative : forall x y z, ((x `union` y) `union` z) = (x `union` y `union` z).
Proof.
 intros; fsetdec.
Qed.

Lemma inter_notin : forall a B, 
  singleton a `inter` B = {} -> a `notin` B.
Proof.
  settac.
Qed.

(*
Lemma notin_inter : forall a B, 
  a `notin` B -> singleton a `inter` B = {}.
Proof.
  intros. fsetdec.
Qed.
*)

Lemma exclude_middle_empty_sets : forall (A:atoms), (A={}) \/ (A <> {}).
Proof.
  intro. pose proof is_empty_total A. destruct H.
    left. pose proof AtomSet.F.is_empty_2 H. fsetdec.
    right. intro. subst. assert (AtomSet.F.is_empty {} = true) ; auto. rewrite H in H0. inversion H0.
Qed.

(* ********************************************************************** *)
(* Utilities about lists *)
(* ********************************************************************** *)

Lemma map_concat_injection : forall (G1:gamma) (G2:gamma) f g, 
  Environment.map f G1 ++ Environment.map g G2 = G1 ++ G2 -> 
  ((Environment.map f G1 = G1) /\ (Environment.map g G2 = G2)).
Admitted.

Lemma list_concat_injection_suffix : forall (A:Type) (G1 : list A) G2 G3, 
  (G1 ++ G2 = G1 ++ G3) -> G2 = G3.
Proof.
  intros. induction G1.
    trivial.
    injection H. assumption.
Qed.

(* ********************************************************************** *)
(* Utilities about ptype *)
(* ********************************************************************** *)

Lemma ptype_cases : forall t A fx,
  (ptype t A fx = t \/ ptype t A fx = typ_qtype t).
Proof.
  intros. unfold ptype. destruct (AtomSet.F.is_empty (A `union` fx)) ; auto.
Qed.

Lemma ptype_union_equality : forall t A1 fx1 A2 fx2,
  (AtomSet.F.is_empty (A1 `union` fx1) = AtomSet.F.is_empty (A2 `union` fx2)) ->
   ptype t A1 fx1 = ptype t A2 fx2.
Proof.
  intros. unfold ptype. rewrite H. reflexivity.
Qed.

Lemma subst_ptype : forall x v t A fx,
  (subst_typ x v (ptype t A fx)) = (ptype (subst_typ x v t) A fx).
Proof.
  intros. unfold ptype. destruct (AtomSet.F.is_empty (A `union` fx)).
    reflexivity.
    simpl. reflexivity.
Qed.

Lemma subst_typ_ptype : forall a b t A fx, 
  (subst_typ_typ a b (ptype t A fx)) = (ptype (subst_typ_typ a b t) A fx).
Proof.
  intros. unfold ptype. destruct (AtomSet.F.is_empty (A `union` fx)).
    reflexivity.
    simpl. reflexivity.
Qed.


(* ********************************************************************** *)
(* Utilities about substitution *)
(* ********************************************************************** *)

Lemma subst_rc_identity :
  (forall e G x v,
    rc_exp e G -> x `notin` dom G -> subst_exp x v e = e) /\
  (forall t G x v,
    rc_typ t G -> x `notin` dom G -> subst_typ x v t = t) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof.
  apply exp_typ_mutual_ind ; intros ; simpl ; trivial ;
  try(inversion H1 ; try(rewrite (H G x v) ; trivial) ; rewrite (H0 G x v) ; trivial) ;
  try(inversion H0 ; rewrite (H G x v) ; trivial).
  destruct (x==a).
    subst. inversion H. pose proof (binds_In _ _ _ _ H2). contradiction.
    reflexivity.
Qed.

(* The existence of [subst_typ_eort] complicates matters, so we need to account
   for [exp_or_typ]s, too. *)
Lemma subst_rc_identity2 :
  (forall e G x v,
    rc_exp e G -> x `notin` dom G -> subst_typ_exp x v e = e) /\
  (forall t G x v,
    rc_typ t G -> x `notin` dom G -> subst_typ_typ x v t = t) /\
  (kind -> True) /\ (*(exp_or_typ -> True).*)
  (forall eort G x v,
    (match eort with
       | eort_exp e =>
         rc_exp e G -> x `notin` dom G -> subst_typ_exp x v e = e
       | eort_typ t =>
         rc_typ t G -> x `notin` dom G -> subst_typ_typ x v t = t
    end)).
Proof.
  apply exp_typ_mutual_ind2 with
    (Q := fun d => forall G x v, rc_domain d G -> x `notin` dom G -> map (subst_typ_eort x v) d = d) ;
  intros ; simpl ;  trivial ;
  try(inversion H1 ; try(rewrite (H G x v) ; trivial) ; rewrite (H0 G x v) ; trivial) ;
  try(inversion H0 ; rewrite (H G x v) ; trivial) ;
  try(apply (H G) ; trivial).
  Case "typ_ftvar". destruct (x==a).
    subst. inversion H. pose proof (binds_In _ _ _ _ H2). contradiction.
    reflexivity.
  Case "eort_exp". subst. simpl. rewrite (H G) ; trivial.
  Case "eort_typ". subst. simpl. rewrite (H G) ; trivial.
Qed.

Lemma subst_typ_wf_exp : forall G e a t,
  wf_exp e G -> a `notin` dom G -> subst_typ_exp a t e = e.
Proof.
  intros. destruct H. eapply (proj1 subst_rc_identity2) ; eassumption.
Qed.

Lemma subst_wf_typ_ext : forall G t x v, 
  wf_typ t G -> x `notin` dom G -> (subst_typ x v t) = t.
Proof.
  intros. destruct H. eapply (proj1 (proj2 subst_rc_identity)) ; eassumption.
Qed.

Lemma subst_wf_exp_ext : forall G e x v, 
  wf_exp e G -> x `notin` dom G -> (subst_exp x v e) = e.
Proof.
  intros. destruct H. eapply (proj1 subst_rc_identity) ; eassumption.
Qed.
(*
(* This lemma is not used anywhere *)
Lemma subst_wf_exp : forall G e x v, 
  wf_exp e G -> wf_exp v G -> wf_exp (subst_exp x v e) G.
Admitted.
*)

Module LAIR_Util (Import X : LAIR_PARAMS).

  Lemma subst_wf_exp_typing : forall G A x e t fx v, 
    wf_exp e nil -> typing G A e (t, fx) -> (subst_typ x v t = t).
  Admitted.

  Lemma subst_wf_exp_typing_on : forall G G' A x e t fx v, 
    only_names G' {} -> wf_exp e G' -> typing G A e (t, fx) -> (subst_typ x v t = t).
  Admitted.

End LAIR_Util.

(* For the facts about substition and opening commuting, I followed *)
(* the structure of the same lemmas and proofs in the file *)
(* Fsub_Infrastructure. *)
Lemma open_expexp_lc_aux :
  (forall e j V i U, i <> j ->
  open_exp j V e = open_exp i U (open_exp j V e) -> e = open_exp i U e) /\
  (forall t j V i U, i <> j ->
  open_typ j V t = open_typ i U (open_typ j V t) -> t = open_typ i U t) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof with eauto*.
  apply exp_typ_mutual_ind; trivial; intros; simpl in *; match goal with H:_=_ |- _ => inversion H end; f_equal...
  Case "typ_bvar".
    destruct (j === n)... destruct (i === n)...
Qed.

Lemma open_typtyp_lc_aux :
  (forall e j V i U, i <> j ->
  open_exp_typ j V e = open_exp_typ i U (open_exp_typ j V e) -> e = open_exp_typ i U e) /\
  (forall t j V i U, i <> j ->
  open_typ_typ j V t = open_typ_typ i U (open_typ_typ j V t) -> t = open_typ_typ i U t) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof with eauto*.
  apply exp_typ_mutual_ind; trivial; intros; simpl in *; match goal with H:_=_ |- _ => inversion H end; f_equal...
  Case "typ_bvar".
    destruct (j === n)... destruct (i === n)...
Qed.

Lemma open_exptyp_lc_aux :
  (forall e j V i U,
  open_exp_typ j V e = open_exp i U (open_exp_typ j V e) -> e = open_exp i U e) /\
  (forall t j V i U,
  open_typ_typ j V t = open_typ i U (open_typ_typ j V t) -> t = open_typ i U t) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof with eauto.
  apply exp_typ_mutual_ind; trivial; intros; simpl in *; match goal with H:_=_ |- _ => inversion H end; f_equal...
Qed.

Lemma open_typexp_lc_aux :
  (forall e j V i U,
  open_exp j V e = open_exp_typ i U (open_exp j V e) -> e = open_exp_typ i U e) /\
  (forall t j V i U,
  open_typ j V t = open_typ_typ i U (open_typ j V t) -> t = open_typ_typ i U t) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof with eauto*.
  apply exp_typ_mutual_ind; trivial; intros; simpl in *; match goal with H:_=_ |- _ => inversion H end; f_equal...
Qed.

Scheme lc_exp_ind2 :=
  Induction for lc_exp Sort Prop
with lc_typ_ind2 :=
  Induction for lc_typ Sort Prop.

Combined Scheme lc_mutind from lc_exp_ind2,lc_typ_ind2.

Lemma open_exp_lc_ident :
  (forall e, lc_exp e -> forall U k, e = open_exp k U e) /\
  (forall t, lc_typ t -> forall U k, t = open_typ k U t).
Proof.
  apply lc_mutind ; intros ; simpl ; f_equal ; auto ; pick fresh x ;
    [eapply (proj1 open_expexp_lc_aux e 0)
      | eapply (proj1 open_exptyp_lc_aux e 0)
      | eapply (proj1 (proj2 open_expexp_lc_aux) t2 0)
      | eapply (proj1 (proj2 open_exptyp_lc_aux) t1 0)];
    eauto.
Qed.

Lemma open_typ_lc_ident :
  (forall e, lc_exp e -> forall T k, e = open_exp_typ k T e) /\
  (forall t, lc_typ t -> forall T k, t = open_typ_typ k T t).
Proof.
  apply lc_mutind ; intros ; simpl ; f_equal ; auto ; pick fresh x ;
    [eapply (proj1 open_typexp_lc_aux e 0)
      | eapply (proj1 open_typtyp_lc_aux e 0)
      | eapply (proj1 (proj2 open_typexp_lc_aux) t2 0)
      | eapply (proj1 (proj2 open_typtyp_lc_aux) t1 0)];
    eauto.
Qed.

Lemma subst_exp_open_exp_aux :
  (forall e x e1 e2 k,
    lc_exp e2 ->
    (subst_exp x e2 (open_exp k e1 e)) = (open_exp k (subst_exp x e2 e1) (subst_exp x e2 e))) /\
  (forall t x e1 e2 k,
    lc_exp e2 ->
    (subst_typ x e2 (open_typ k e1 t)) = (open_typ k (subst_exp x e2 e1) (subst_typ x e2 t))) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof.
  apply exp_typ_mutual_ind ; intros ; simpl ; f_equal ; auto.
    Case "exp_bvar".
    destruct (k===n) ; subst ; trivial.
    Case "exp_fvar".
    destruct (x==a) ; subst ; trivial. apply (proj1 open_exp_lc_ident). assumption.
Qed.

Lemma subst_exp_open_exp_typ_aux :
  (forall e x T E k,
    lc_exp E ->
    (subst_exp x E (open_exp_typ k T e)) = (open_exp_typ k (subst_typ x E T) (subst_exp x E e))) /\
  (forall t x T E k,
    lc_exp E ->
    (subst_typ x E (open_typ_typ k T t)) = (open_typ_typ k (subst_typ x E T) (subst_typ x E t))) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof.
  apply exp_typ_mutual_ind ; intros ; simpl ; f_equal ; auto.
    Case "exp_fvar".
    destruct (x==a) ; subst ; trivial. apply (proj1 open_typ_lc_ident). assumption.
    Case "typ_btvar".
    destruct (k===n) ; subst ; trivial.
Qed.

Lemma subst_typ_exp_open_exp_typ_aux :
  (forall e x T T1 k,
    lc_typ T1 ->
    (subst_typ_exp x T1 (open_exp_typ k T e)) = (open_exp_typ k (subst_typ_typ x T1 T) (subst_typ_exp x T1 e))) /\
  (forall t x T T1 k,
    lc_typ T1 ->
    (subst_typ_typ x T1 (open_typ_typ k T t)) = (open_typ_typ k (subst_typ_typ x T1 T) (subst_typ_typ x T1 t))) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof.
  apply exp_typ_mutual_ind ; intros ; simpl ; f_equal ; auto.
    Case "typ_bvar".
    destruct (k===n) ; subst ; trivial.
    Case "typ_ftvar".
    destruct (x==a) ; subst ; trivial. apply (proj2 open_typ_lc_ident). assumption.
Qed.

Lemma subst_typ_exp_open_exp_aux :
  (forall e x E T k,
    lc_typ T ->
    (subst_typ_exp x T (open_exp k E e)) = (open_exp k (subst_typ_exp x T E) (subst_typ_exp x T e))) /\
  (forall t x E T k,
    lc_typ T ->
    (subst_typ_typ x T (open_typ k E t)) = (open_typ k (subst_typ_exp x T E) (subst_typ_typ x T t))) /\
  (kind -> True) /\ (exp_or_typ -> True).
Proof.
  apply exp_typ_mutual_ind ; intros ; simpl ; f_equal ; auto.
    Case "typ_bvar".
    destruct (k===n) ; subst ; trivial.
    Case "exp_fvar".
    destruct (x==a) ; subst ; trivial. apply (proj2 open_exp_lc_ident). assumption.
Qed.

Lemma subst_exp_open_exp_commute : forall (x y : atom) vx e, 
  x <> y -> lc_exp vx -> 
  (subst_exp x vx (open_exp 0 y e)) = (open_exp 0 y (subst_exp x vx e)).
Proof.
  destruct subst_exp_open_exp_aux as [H _]. intros x y vx e neq. assert (subst_exp x vx y = y) as Heq by (simpl ; destruct (x==y) ; [contradiction | reflexivity]). pose proof (H e x y vx 0) as hyp. rewrite Heq in hyp. assumption.
Qed.

Lemma subst_typ_open_typ_commute : forall (x y : atom) vx t, 
  x <> y -> lc_exp vx -> 
  (subst_typ x vx (open_typ 0 y t)) = (open_typ 0 y (subst_typ x vx t)).
Proof.
  destruct subst_exp_open_exp_aux as [_ [H _]]. intros x y vx e neq. assert (subst_exp x vx y = y) as Heq by (simpl ; destruct (x==y) ; [contradiction | reflexivity]). pose proof (H e x y vx 0) as hyp. rewrite Heq in hyp. assumption.
Qed.

(* For the next two, we don't need the assumption [x <> b] because exp variables
   are different from typ variables. *)
Lemma subst_typ_open_typ_typ_commute : forall (x b : atom) v t,
  lc_exp v ->
  (subst_typ x v (open_typ_typ 0 b t) = open_typ_typ 0 b (subst_typ x v t)).
Proof.
  destruct subst_exp_open_exp_typ_aux as [_ [H _]]. intros x y vx e. apply (H e x y vx).
Qed.

Lemma subst_exp_open_exp_typ_commute : forall (x b : atom) v e,
  lc_exp v -> 
  (subst_exp x v (open_exp_typ 0 b e) = open_exp_typ 0 b (subst_exp x v e)).
Proof.
  destruct subst_exp_open_exp_typ_aux as [H _]. intros x y vx e. apply (H e x y vx).
Qed.

(* [y] must be an atom for this to be true. Without specifying this,
   [y] is inferred to be an [exp], but then the statement is not true,
   as the next (commented out) lemma illustrates. *)
Lemma subst_typ_exp_open_exp_commute : forall a b (y:atom) e, 
  lc_typ b -> 
  (subst_typ_exp a b (open_exp 0 y e)) = (open_exp 0 y (subst_typ_exp a b e)).
Proof.
  destruct subst_typ_exp_open_exp_aux as [H _]. intros x T E e. apply (H e x E T).
Qed.

(*
(* Why [y] has to be restricted to being an atom *)
Lemma subst_typ_exp_open_exp_commute : ~forall a b y e, 
  lc_typ b -> 
  (subst_typ_exp a b (open_exp 0 y e)) = (open_exp 0 y (subst_typ_exp a b e)).
Proof.
  intro. set (e := exp_bvar 0). pick fresh a. set (y := exp_lam (typ_ftvar a) exp_bot). pick fresh x.
assert (lc_typ x) by constructor. pose proof H a (typ_ftvar x) y e H0.
simpl in H1. destruct (a==a). subst. injection H1. intro. subst. fsetdec.
destruct n. reflexivity.
Qed.
*)

Lemma open_typ_subst_typ_typ_commute : forall a b (y:atom) t,
  lc_typ b -> 
  (open_typ 0 y (subst_typ_typ a b t)) = (subst_typ_typ a b (open_typ 0 y t)).
Proof.
  destruct subst_typ_exp_open_exp_aux as [_ [H _]]. intros x T E t lc. symmetry. apply (H t x E T). assumption.
Qed.

Lemma open_exp_typ_subst_typ_exp_commute : forall a (b:atom) e (c:atom), 
  lc_typ b -> a <> c ->
  (open_exp_typ 0 c (subst_typ_exp a b e)) = (subst_typ_exp a b (open_exp_typ 0 c e)).
Proof.
  destruct subst_typ_exp_open_exp_typ_aux as [H _]. intros x T1 e T lc neq. symmetry. assert (subst_typ_typ x T1 T = T) as Heq by (simpl ; destruct (x==T) ; [contradiction | reflexivity]). pose proof (H e x T T1 0 lc) as hyp. rewrite Heq in hyp. assumption.
Qed.

Lemma open_typ_typ_subst_typ_typ_commute : forall a (b:atom) t (c:atom), 
  lc_typ b -> a <> c ->
  (open_typ_typ 0 c (subst_typ_typ a b t)) = (subst_typ_typ a b (open_typ_typ 0 c t)).
Proof.
  destruct subst_typ_exp_open_exp_typ_aux as [_ [H _]]. intros x T1 e T lc neq. symmetry. assert (subst_typ_typ x T1 T = T) as Heq by (simpl ; destruct (x==T) ; [contradiction | reflexivity]). pose proof (H e x T T1 0 lc) as hyp. rewrite Heq in hyp. assumption.
Qed.

Lemma subst_typ_open_typ_commute_weak : forall x v e t, 
  lc_exp v -> 
  (subst_typ x v (open_typ 0 e t)) = (open_typ 0 (subst_exp x v e) (subst_typ x v t)).
Proof.
  destruct subst_exp_open_exp_aux as [_ [H _]]. intros x v e t. apply H.
Qed.

Lemma subst_typ_open_typ_typ_commute_weak : forall x v t t',
  lc_exp v ->
  subst_typ x v (open_typ_typ 0 t t') = open_typ_typ 0 (subst_typ x v t) (subst_typ x v t').
Proof.
  destruct subst_exp_open_exp_typ_aux as [_ [H _]]. intros x v e t. apply H.
Qed.

Lemma subst_typ_typ_open_typ_commute_weak : forall a b e t,
  lc_typ b ->
  subst_typ_typ a b (open_typ 0 e t) = open_typ 0 (subst_typ_exp a b e) (subst_typ_typ a b t).
Proof.
  destruct subst_typ_exp_open_exp_aux as [_ [H _]]. intros x v e t. apply H.
Qed.

(*
(* Substitution assumes that its second argument is locally closed.
   Without local closure, things fall apart. Consider this example,
   where the previous lemma fails without local closure. *)
Lemma one: forall x, subst_typ_typ x 0 (open_typ_typ 0 1 x) = 0.
Proof.
  intros. simpl. destruct (x==x). trivial. destruct n. trivial.
Qed.
Lemma two: forall x, open_typ_typ 0 (subst_typ_typ x 0 1) (subst_typ_typ x 0 x) = 1.
  intros. simpl. destruct (x==x). simpl. trivial. contradiction n. trivial.
Qed.

Lemma bad_subst_typ_typ_open_typ_typ_commute_weak : ~forall a b t t',
  (subst_typ_typ a b (open_typ_typ 0 t t')) = (open_typ_typ 0 (subst_typ_typ a b t) (subst_typ_typ a b t')).
Proof.
  intro. pick fresh x. pose proof H x 0 1 x.
  rewrite one in H0. rewrite two in H0. simpl in H0. discriminate.
Qed.
*)

Lemma subst_typ_typ_open_typ_typ_commute_weak : forall a b t t',
  lc_typ b ->
  (subst_typ_typ a b (open_typ_typ 0 t t')) = (open_typ_typ 0 (subst_typ_typ a b t) (subst_typ_typ a b t')).
Proof.
  destruct subst_typ_exp_open_exp_typ_aux as [_ [H _]]. intros x T1 T t. apply H.
Qed.

Lemma subst_exp_intro : forall G x e v k, 
  rc_exp e G -> x `notin` dom G -> (open_exp k v e = (subst_exp x v (open_exp k x e))).
Admitted.

Lemma subst_typ_intro : forall G x t v k, 
  rc_typ t G -> x `notin` dom G -> (open_typ k v t = (subst_typ x v (open_typ k x t))).
Admitted.

Lemma subst_typ_exp_intro : forall G t e b k,  
 rc_exp e G -> b `notin` dom G -> (open_exp_typ k t e = (subst_typ_exp b t (open_exp_typ k b e))).
Admitted.

Lemma subst_typ_typ_intro : forall G t t' b k, 
  rc_typ t' G -> b `notin` dom G -> (open_typ_typ k t t' = (subst_typ_typ b t (open_typ_typ k b t'))).
Admitted.

(* ********************************************************************** *)
(* Utilities for aqual *)
(* ********************************************************************** *)

Lemma aqual_cases : forall A x k, 
  (aqual x k A) = A  \/ (aqual x k A) = (singleton x `union` A).
Admitted.

Lemma aqual_identity : forall X x k A, 
  X `in` aqual x k A -> (X = x \/ X `in` A).
Admitted.

Lemma aqual_identity_0 : forall X x k A, 
  X `in` A -> X `in` (aqual x k A).
Admitted.

Lemma aqual_union_identity_1 : forall X A0 x k A, 
  (singleton X `union` A0) = (aqual x k A) -> (X `notin` A0) -> (x `notin` A) -> 
  exists A1, (A0 = aqual x k A1 /\ x `notin` A1).
Admitted.

Lemma aqual_union_identity_2 : forall X x k A1 A, 
  (singleton X `union` A1 = aqual x k A) -> 
  (x <> X) ->
  exists A2, (A = singleton X `union` A2 /\ X `notin` A2).
Admitted.

Lemma aqual_identity_3 : forall X x k A A1 A2, 
  X `notin` A1 -> X `notin` A2 -> x `notin` A1 -> x `notin` A -> 
  (singleton X `union` aqual x k A1) = aqual x k A -> 
  (singleton X `union` A2) = A -> A1 = A2.
Admitted.

Lemma Ax_empty_A_empty : forall A x k, (aqual x k A) = {} -> A = {}.
Admitted.

Lemma aqual_commutes : forall A x k_x y k_y, 
  aqual x k_x (aqual y k_y A) = aqual y k_y (aqual x k_x A).
Admitted.
(*
(* These 2 are not used anywhere *)
Lemma aqual_split_commute : forall A A0 x k a, 
  splitA (singleton a) A (aqual x k A0) ->  x <> a -> 
  exists A1, A = aqual x k A1 /\ (x `notin` A1).
Admitted.

Lemma split_aqual_dom_containment : forall A0 x k A A', 
  splitA A0 (aqual x k A) (aqual x k A') -> x `notin` A -> AtomSet.F.Subset A A'.
Admitted.
*)
